# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib._parent**
"""

import sys
import textwrap
import warnings

from pluginlib import (abstractmethod, abstractproperty, abstractstaticmethod,
                       abstractclassmethod, abstractattribute)
import pluginlib._parent as parent

from tests import OUTPUT, TestCase, unittest


NO_ASYNC = sys.version_info[:2] < (3, 5)
PY2 = sys.version_info[0] < 3


# pylint: disable=protected-access, no-member


# This should be imported from types, but it's broken in pypy
# https://bitbucket.org/pypy/pypy/issues/2865
class Placeholder(object):
    """Placeholder to get type"""
    __slots__ = ('member',)


MemberDescriptorType = type(Placeholder.__dict__['member'])  # pylint: disable=invalid-name


class TestParent(TestCase):
    """Test Parent decorator"""

    def tearDown(self):

        # Clear plugins from module
        parent.get_plugins().clear()

    def test_parent_basic(self):
        """Basic use of Parent decorator"""

        @parent.Parent('basic')
        class Basic(object):
            """Basic document string"""

        self.assertTrue(issubclass(Basic, parent.Plugin))
        self.assertTrue(isinstance(Basic, parent.PluginType))
        self.assertTrue(Basic._parent_)
        self.assertTrue(Basic._skipload_)
        self.assertEqual(Basic.__doc__, 'Basic document string')
        self.assertEqual(Basic._type_, 'basic')
        self.assertTrue('basic' in parent.get_plugins()['_default'])

    def test_parent_no_type(self):
        """Parent type defaults to parent class name"""

        @parent.Parent()
        class Basic(object):
            """Basic document string"""

        self.assertEqual(Basic._type_, 'Basic')
        self.assertTrue('Basic' in parent.get_plugins()['_default'])

    def test_parent_bare(self):
        """Parent type defaults to parent class name"""

        @parent.Parent
        class Basic(object):
            """Basic document string"""

        self.assertEqual(Basic._type_, 'Basic')
        self.assertTrue('Basic' in parent.get_plugins()['_default'])

    def test_multiple_inheritance(self):
        """Base class is inherited from another class"""

        class Sample(object):
            """Sample Class"""

        @parent.Parent('multiple_inheritence')
        class MultiInheritance(Sample):
            """TestMultiInheritance"""

        self.assertTrue(issubclass(MultiInheritance, parent.Plugin))
        self.assertTrue(isinstance(MultiInheritance, parent.PluginType))
        self.assertTrue(issubclass(MultiInheritance, Sample))
        self.assertTrue(MultiInheritance._parent_)
        self.assertTrue(MultiInheritance._skipload_)

    def child_is_parent(self, peers=False):
        """
        A parent class is used as base class for new parent
        If peers is True, another class is created with the same alias
        """

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent"""

            def hello(self):  # pylint: disable=no-self-use
                """Hello, world!"""
                return 'world'

        if peers:

            class Child(Parent):
                """Child"""

                _alias_ = 'foo'
                _version_ = '1.0'

        @parent.Parent('child_is_parent')
        class ParentChild(Parent):
            """TestParentChild"""

            _alias_ = 'foo'

        self.assertTrue(issubclass(ParentChild, parent.Plugin))
        self.assertTrue(isinstance(ParentChild, parent.PluginType))
        self.assertTrue(issubclass(ParentChild, Parent))
        self.assertTrue(ParentChild._parent_)
        self.assertTrue(ParentChild._skipload_)
        self.assertTrue(hasattr(ParentChild, 'hello'))

        if peers:
            self.assertTrue(len(Parent._get_plugins()['foo']) == 1)
            self.assertTrue(Parent._get_plugins()['foo']['1.0'] is Child)

        else:
            self.assertFalse('foo' in Parent._get_plugins())

        class GrandChild(ParentChild):
            """GrandChild"""

        self.assertTrue(issubclass(GrandChild, Parent))
        self.assertTrue(issubclass(GrandChild, ParentChild))
        self.assertTrue(hasattr(GrandChild, 'hello'))

    def test_child_is_parent(self):
        """A parent class is used as base class for new parent"""
        self.child_is_parent(False)

    def test_child_is_parent_peers(self):
        """A parent class is used as base class for new parent, has peers"""
        self.child_is_parent(True)

    def test_slots(self):
        """slots in a parent work as expected"""

        @parent.Parent
        class WithSlots(object):
            """This class has slots"""
            __slots__ = ('ivar',)

            def __init__(self, ivar):
                self.ivar = ivar

        inst = WithSlots(1)

        self.assertTrue('ivar' in WithSlots.__dict__)
        # Usually slots are member descriptors, but in pypy they are getset descriptors
        self.assertIsInstance(WithSlots.__dict__['ivar'], MemberDescriptorType,)
        self.assertTrue(hasattr(inst, '__slots__'))
        self.assertFalse(hasattr(inst, '__dict__'))
        self.assertEqual(inst.ivar, 1)


class TestPlugin(TestCase):
    """Test Plugin mixin"""

    def setUp(self):

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent"""

            def hello(self):  # pylint: disable=no-self-use
                """Hello, world!"""
                return 'world'

        self.parent = Parent
        self.plugins = self.parent._get_plugins()

    def tearDown(self):

        # Clear plugins from module
        parent.get_plugins().clear()

    def test_plugin_type(self):
        """plugin_type property is passed from parent"""
        self.assertEqual(self.parent.plugin_type, 'test_parent')

        class Child(self.parent):
            """Child has same plugin_type"""

        self.assertEqual(Child.plugin_type, 'test_parent')

    def test_plugin_group_none(self):
        """plugin_group property is passed from parent"""
        self.assertEqual(self.parent.plugin_group, None)

        class Child(self.parent):
            """Child has same plugin_group"""

        self.assertEqual(Child.plugin_group, None)

    def test_plugin_group(self):
        """plugin_group property is passed from parent"""

        @parent.Parent('test_parent', group='A-Team')
        class Parent(object):
            """Parent"""

        self.assertEqual(Parent.plugin_group, 'A-Team')

        class Child(Parent):
            """Child has same plugin_group"""

        self.assertEqual(Child.plugin_group, 'A-Team')

    def test_alias_none(self):
        """name attribute is class name"""

        class Child(self.parent):
            """No alias"""

        self.assertTrue(issubclass(Child, self.parent))
        self.assertEqual(Child.name, 'Child')
        self.assertIs(self.plugins['Child']['0'], Child)

    def test_alias(self):
        """name attribute is _alias_"""

        class Child(self.parent):
            """With alias"""
            _alias_ = 'voldemort'

        self.assertTrue(issubclass(Child, self.parent))
        self.assertEqual(Child.name, 'voldemort')
        self.assertIs(self.plugins['voldemort']['0'], Child)
        self.assertFalse('Child' in self.plugins)

    def test_version_none(self):
        """version attribute is _version_, module __version__ or None"""

        class Child(self.parent):
            """No version"""
            _alias_ = 'the_boy_who_lived'

        self.assertEqual(Child.version, None)
        self.assertIs(self.plugins['the_boy_who_lived']['0'], Child)

    def test_version_class(self):
        """version attribute is _version_"""

        class Child(self.parent):
            """With version"""
            _version_ = '1.0.1'
            _alias_ = 'the_boy_who_lived'

        self.assertEqual(Child.version, '1.0.1')
        self.assertIs(self.plugins['the_boy_who_lived']['1.0.1'], Child)

    def test_version_module(self):
        """version attribute is _version_, module __version__ or None"""

        sys.modules[__name__].__version__ = '2.0.0'

        class Child(self.parent):
            """Module version"""
            _alias_ = 'the_boy_who_lived'

        self.assertEqual(Child.version, '2.0.0')
        self.assertIs(self.plugins['the_boy_who_lived']['2.0.0'], Child)

        delattr(sys.modules[__name__], '__version__')

    def test_skipload(self):
        """Child not loaded when _skipload_ is True"""

        class Child(self.parent):  # pylint: disable=unused-variable
            """_skipload_ is True"""
            _skipload_ = True

        self.assertFalse('Child' in self.plugins)


# pylint: disable=too-many-public-methods
class TestPluginType(TestCase):
    """Test PluginType metaclass"""

    def setUp(self):

        # Truncate log output
        OUTPUT.seek(0)
        OUTPUT.truncate(0)

    def tearDown(self):

        # Clear plugins from module
        parent.get_plugins().clear()

    def test_skipload_static(self):
        """Use _skipload_ static method to determine if plugin should be used"""

        skip = False

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with callable _skipload_"""

        plugins = Parent._get_plugins()

        class Child1(Parent):
            """Child succeeds"""

            @staticmethod
            def _skipload_():
                """Don't skip"""
                return skip

        self.assertIs(plugins['Child1']['0'], Child1)
        self.assertEqual(OUTPUT.getvalue(), '')

        skip = True, 'This parrot is no more!'

        class Child2(Parent):  # pylint: disable=unused-variable
            """Child fails"""

            @staticmethod
            def _skipload_():
                """Skip"""
                return skip

        self.assertFalse('Child2' in plugins)
        self.assertRegex(OUTPUT.getvalue(), 'This parrot is no more!')

    def test_skipload_class(self):
        """Use _skipload_ class method to determine if plugin should be used"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with callable _skipload_"""

        plugins = Parent._get_plugins()

        class Child1(Parent):
            """Child succeeds"""

            skip = False

            @classmethod
            def _skipload_(cls):
                """Don't skip"""
                return cls.skip

        self.assertIs(plugins['Child1']['0'], Child1)
        self.assertEqual(OUTPUT.getvalue(), '')

        class Child2(Parent):  # pylint: disable=unused-variable
            """Child fails"""

            skip = True, 'He has ceased to be!'

            @classmethod
            def _skipload_(cls):
                """Don't skip"""
                return cls.skip

        self.assertFalse('Child2' in plugins)
        self.assertRegex(OUTPUT.getvalue(), 'He has ceased to be!')

    def test_abstract_method(self):
        """Method required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract class method"""

            @abstractmethod
            def abstract(self):
                """Abstract method"""

        self.missing(Parent, 'Does not contain required method')
        self.meth(Parent)
        self.static(Parent, 'Does not contain required method')
        self.klass(Parent, 'Does not contain required method')
        self.prop(Parent, 'Does not contain required method')
        self.attr(Parent, 'Does not contain required method')
        self.coroutine(Parent)
        self.type_hints_1(Parent)
        self.type_hints_2(Parent)

    def test_multiple_methods(self):
        """Method required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract class method"""

            @abstractmethod
            def abstract2(self):
                """Abstract method"""

            @abstractmethod
            def abstract(self):
                """Abstract method"""

        self.meth(Parent, 'Does not contain required method')
        self.multiple(Parent)

    def test_abs_method_argspec(self):
        """Method argument spec must match"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract class method"""

            @abstractmethod
            def abstract(self, some_arg):
                """Abstract method with another argument"""

        self.meth(Parent, 'Argument spec does not match parent for method')

    def test_abstract_staticmethod(self):
        """Static method required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract static method"""

            @abstractstaticmethod
            def abstract():
                """Abstract static method"""

        self.missing(Parent, 'Does not contain required static method')
        self.meth(Parent, 'Does not contain required static method')
        self.static(Parent)
        self.klass(Parent, 'Does not contain required static method')
        self.prop(Parent, 'Does not contain required static method')
        self.attr(Parent, 'Does not contain required static method')

    def test_abs_static_meth_argspec(self):
        """Static method argument spec must match"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract static method"""

            @abstractstaticmethod
            def abstract(some_arg):
                """Abstract static method with another argument"""

        self.static(Parent, 'Argument spec does not match parent for method')

    def test_abstract_classmethod(self):
        """Class method required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract class method"""

            @abstractclassmethod
            def abstract(cls):  # noqa: N805
                """Abstract class method"""

        self.missing(Parent, 'Does not contain required class method')
        self.meth(Parent, 'Does not contain required class method')
        self.static(Parent, 'Does not contain required class method')
        self.klass(Parent)
        self.prop(Parent, 'Does not contain required class method')
        self.attr(Parent, 'Does not contain required class method')

    def test_abs_class_meth_argspec(self):
        """Class method argument spec must match"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract class method"""

            @abstractclassmethod
            def abstract(cls, some_arg):  # noqa: N805
                """Abstract class method with another argument"""

        self.klass(Parent, 'Argument spec does not match parent for method')

    def test_abstract_property(self):
        """Property required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract property"""

            @abstractproperty
            def abstract(self):
                """Abstract property"""

        self.missing(Parent, 'Does not contain required property')
        self.meth(Parent, 'Does not contain required property')
        self.static(Parent, 'Does not contain required property')
        self.klass(Parent, 'Does not contain required property')
        self.prop(Parent)
        self.attr(Parent, 'Does not contain required property')

    def test_abstract_attribute(self):
        """Attribute required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract attribute"""

            abstract = abstractattribute

        self.missing(Parent, 'Does not contain required attribute')
        self.meth(Parent)
        self.static(Parent)
        self.klass(Parent)
        self.prop(Parent)
        self.attr(Parent)

    @unittest.skipIf(NO_ASYNC, 'Requires Python 3.5+')
    def test_abstract_coroutine(self):
        """Attribute required in subclass"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with abstract coroutine"""

            class_definition = textwrap.dedent('''\
            @abstractmethod
            async def abstract(self):
                """Abstract coroutine method"""
            ''')

            local_rtn = {}
            exec(class_definition, globals(), local_rtn)  # pylint: disable=exec-used
            abstract = local_rtn['abstract']

        self.missing(Parent, 'Does not contain required method')
        self.meth(Parent, 'Does not contain required coroutine method')
        self.static(Parent, 'Does not contain required method')
        self.klass(Parent, 'Does not contain required method')
        self.prop(Parent, 'Does not contain required method')
        self.attr(Parent, 'Does not contain required method')
        self.coroutine(Parent)

    @unittest.skipIf(PY2, 'Requires Python 3+')
    def test_type_annotations(self):
        """If parent and child have annotations they must match"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent with typed method"""

            class_definition = textwrap.dedent('''\
            @abstractmethod
            def abstract(self) -> str:
                """Method with type annotations"""
            ''')

            local_rtn = {}
            exec(class_definition, globals(), local_rtn)  # pylint: disable=exec-used
            abstract = local_rtn['abstract']

        self.meth(Parent)
        self.type_hints_1(Parent)
        self.type_hints_2(Parent, 'Type annotations differ')

    def check_method(self, parent_class, error, child, e):
        """Check child methods"""

        self.assertEqual(OUTPUT.getvalue(), '')

        plugins = parent_class._get_plugins()

        if error:
            self.assertFalse(child.name in plugins)
            self.assertEqual(len(e), 1)
            self.assertRegex(str(e[-1].message), error)
        else:
            self.assertIs(plugins[child.name]['0'], child)
            self.assertEqual(len(e), 0)

    def missing(self, parent_class, error=None):
        """Test abstract method is missing"""

        with warnings.catch_warnings(record=True) as e:

            class Missing(parent_class):
                """Does not have abstract method"""

        self.check_method(parent_class, error, Missing, e)

    def meth(self, parent_class, error=None):
        """Test abstract method is a regular method"""

        with warnings.catch_warnings(record=True) as e:

            class Meth(parent_class):
                """Only has regular method"""

                def abstract(self):
                    """Regular method"""

        self.check_method(parent_class, error, Meth, e)

    def multiple(self, parent_class, error=None):
        """Test abstract method is a regular method"""

        with warnings.catch_warnings(record=True) as e:

            class Meth(parent_class):
                """Only has regular method"""

                def abstract(self):
                    """Regular method"""

                def abstract2(self):
                    """Regular method"""

        self.check_method(parent_class, error, Meth, e)

    def static(self, parent_class, error=None):
        """Test abstract method is a static method"""

        with warnings.catch_warnings(record=True) as e:

            class Static(parent_class):
                """Only has static method"""

                @staticmethod
                def abstract():
                    """Static method"""

        self.check_method(parent_class, error, Static, e)

    def klass(self, parent_class, error=None):
        """Test abstract method is a class method"""

        with warnings.catch_warnings(record=True) as e:

            class Klass(parent_class):
                """Only has class method"""

                @classmethod
                def abstract(cls):
                    """Class method"""

        self.check_method(parent_class, error, Klass, e)

    def prop(self, parent_class, error=None):
        """Test abstract method is a property"""

        with warnings.catch_warnings(record=True) as e:

            class Prop(parent_class):
                """Only has property"""

                @property
                def abstract(self):
                    """Property"""

                @abstract.setter
                def abstract(self, value):
                    pass

                @abstract.deleter
                def abstract(self):
                    pass

        self.check_method(parent_class, error, Prop, e)

    def attr(self, parent_class, error=None):
        """Test abstract attribute"""

        with warnings.catch_warnings(record=True) as e:

            class Attr(parent_class):
                """Only has property"""
                abstract = 'No one expects the abstract attribute'

        self.check_method(parent_class, error, Attr, e)

    def coroutine(self, parent_class, error=None):
        """Test abstract coroutine method"""

        if NO_ASYNC:
            return

        with warnings.catch_warnings(record=True) as e:

            class_definition = textwrap.dedent('''\
            class Coroutine(parent_class):
                """Only has coroutine method"""
                async def abstract(self):
                    """Coroutine method"""
            ''')

            local_rtn = {'parent_class': parent_class}
            exec(class_definition, globals(), local_rtn)  # pylint: disable=exec-used
            Coroutine = local_rtn['Coroutine']  # pylint: disable=invalid-name

        self.check_method(parent_class, error, Coroutine, e)

    def type_hints_1(self, parent_class, error=None):
        """Test abstract method has type annotations"""

        if PY2:
            return

        with warnings.catch_warnings(record=True) as e:

            class_definition = textwrap.dedent('''\
            class TypeHints1(parent_class):
                """Has type annotations"""

                def abstract(self) -> str:
                    """Method with type annotations"""
            ''')

            local_rtn = {'parent_class': parent_class}
            exec(class_definition, globals(), local_rtn)  # pylint: disable=exec-used
            TypeHints1 = local_rtn['TypeHints1']  # pylint: disable=invalid-name

        self.check_method(parent_class, error, TypeHints1, e)

    def type_hints_2(self, parent_class, error=None):
        """Test abstract method has type annotations"""

        if PY2:
            return

        with warnings.catch_warnings(record=True) as e:

            class_definition = textwrap.dedent('''\
            class TypeHints2(parent_class):
                """Has type annotations"""

                def abstract(self) -> int:
                    """Method with type annotations"""
            ''')

            local_rtn = {'parent_class': parent_class}
            exec(class_definition, globals(), local_rtn)  # pylint: disable=exec-used
            TypeHints2 = local_rtn['TypeHints2']  # pylint: disable=invalid-name

        self.check_method(parent_class, error, TypeHints2, e)

    def test_duplicate_parents(self):
        """Parents with the same plugin type should raise an error"""

        @parent.Parent('test_parent')
        class Parent1(object):  # pylint: disable=unused-variable
            """First Parent"""

        with self.assertRaisesRegex(ValueError, "parent must be unique"):

            @parent.Parent('test_parent')
            class Parent2(object):  # pylint: disable=unused-variable
                """Second Parent"""

    def test_duplicate_versions(self):
        """Duplicate versions throw warning and are ignored"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent"""

        plugins = Parent._get_plugins()

        with warnings.catch_warnings(record=True) as e:

            class Child1(Parent):
                """First Child"""
                _alias_ = 'child'

        self.assertEqual(len(plugins['child']), 1)
        self.assertIs(plugins['child']['0'], Child1)
        self.assertEqual(len(e), 0)

        with warnings.catch_warnings(record=True) as e:

            class Child2(Parent):  # pylint: disable=unused-variable
                """Duplicate Child"""
                _alias_ = 'child'

        self.assertEqual(len(plugins['child']), 1)
        self.assertIs(plugins['child']['0'], Child1)
        self.assertEqual(len(e), 1)
        self.assertRegex(str(e[-1].message), 'Duplicate plugins found')

        with warnings.catch_warnings(record=True) as e:

            class Child3(Parent):
                """Child with a different version"""
                _alias_ = 'child'
                _version_ = '1.0'

        self.assertEqual(len(plugins['child']), 2)
        self.assertIs(plugins['child']['0'], Child1)
        self.assertIs(plugins['child']['1.0'], Child3)
        self.assertEqual(len(e), 0)

    def test_unknown_parent(self):
        """Raise an exception if parent is an unknown type"""

        @parent.Parent('test_parent')
        class Parent(object):
            """Parent"""

        # Clear registry
        parent.get_plugins().clear()

        with self.assertRaisesRegex(ValueError, "Unknown parent type"):

            class Child1(Parent):  # pylint: disable=unused-variable
                """Child"""
                _alias_ = 'child'



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib**
"""

import logging
import sys

from pluginlib._util import LOGGER

if sys.version_info[:2] < (2, 7):
    import unittest2 as unittest  # pylint: disable=import-error
else:
    import unittest  # pylint: disable=wrong-import-order

if sys.version_info[:2] < (3, 3):
    import mock  # pylint: disable=import-error
else:
    from unittest import mock  # noqa: F401  # pylint: disable=no-name-in-module, import-error

if sys.version_info[0] < 3:
    from StringIO import StringIO  # pylint: disable=import-error
else:
    from io import StringIO

OUTPUT = StringIO()
HANDLER = logging.StreamHandler(OUTPUT)
LOGGER.addHandler(HANDLER)
LOGGER.setLevel(logging.INFO)

__all__ = ['mock', 'unittest', 'OUTPUT', 'TestCase']


class TestCase(unittest.TestCase):
    """Simple subclass of unittest.TestCase"""


# Fix deprecated methods
def assert_regex(self, text, regex, msg=None):
    """
    Wrapper for assertRegexpMatches
    """

    return self.assertRegexpMatches(text, regex, msg)


def assert_not_regex(self, text, regex, msg=None):
    """
    Wrapper for assertNotRegexpMatches
    """

    return self.assertNotRegexpMatches(text, regex, msg)


def assert_raises_regex(self, exception, regex, *args, **kwargs):
    """
    Wrapper for assertRaisesRegexp
    """

    return self.assertRaisesRegexp(exception, regex, *args, **kwargs)


if not hasattr(TestCase, 'assertRegex'):
    TestCase.assertRegex = assert_regex

if not hasattr(TestCase, 'assertNotRegex'):
    TestCase.assertNotRegex = assert_not_regex

if not hasattr(TestCase, 'assertRaisesRegex'):
    TestCase.assertRaisesRegex = assert_raises_regex



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib.exceptions**
"""

import warnings

from pluginlib import exceptions

from tests import TestCase


# Because we want the test names to match the class names
# pylint: disable=invalid-name
class TestExceptions(TestCase):
    """Tests for custom exceptions"""

    def setUp(self):
        self.msg = 'Just your friendly neighborhood Spiderman'

    def test_pluginlib_error(self):
        """Raises PluginlibError as expected"""
        with self.assertRaises(Exception) as e:
            raise exceptions.PluginlibError()
        self.assertIsNone(e.exception.friendly)

        with self.assertRaises(Exception) as e:
            raise exceptions.PluginlibError(friendly=self.msg)
        self.assertEqual(e.exception.friendly, self.msg)

    def test_plugin_import_error(self):
        """Raises PluginImportError as expected"""
        with self.assertRaises(exceptions.PluginlibError) as e:
            raise exceptions.PluginImportError()
        self.assertIsNone(e.exception.friendly)

        with self.assertRaises(exceptions.PluginlibError) as e:
            raise exceptions.PluginImportError(friendly=self.msg)
        self.assertEqual(e.exception.friendly, self.msg)


class TestWarnings(TestCase):
    """Tests for custom warnings"""

    def test_plugin_warning(self):
        """Warns with PluginWarning as expected"""

        with warnings.catch_warnings(record=True) as e:
            warnings.simplefilter("always")
            warnings.warn('Just a warning', exceptions.PluginWarning)
            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, exceptions.PluginWarning))
            self.assertTrue(issubclass(e[-1].category, UserWarning))
            self.assertEqual(str(e[-1].message), 'Just a warning')

    def test_entrypoint_warning(self):
        """Warns with EntryPointWarning as expected"""

        with warnings.catch_warnings(record=True) as e:
            warnings.simplefilter("always")
            warnings.warn('Just a warning', exceptions.EntryPointWarning)
            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, exceptions.EntryPointWarning))
            self.assertTrue(issubclass(e[-1].category, ImportWarning))
            self.assertEqual(str(e[-1].message), 'Just a warning')



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Package**

A framework for creating and importing plugins
"""

__version__ = '0.8.1'

__all__ = ['abstractmethod', 'abstractproperty', 'abstractstaticmethod', 'abstractclassmethod',
           'abstractattribute', 'BlacklistEntry', 'EntryPointWarning', 'Parent', 'Plugin',
           'PluginlibError', 'PluginImportError', 'PluginLoader', 'PluginWarning']

from abc import abstractmethod, abstractproperty

from pluginlib._parent import Parent, Plugin
from pluginlib._loader import PluginLoader
from pluginlib._objects import BlacklistEntry
from pluginlib._util import abstractstaticmethod, abstractclassmethod, abstractattribute
from pluginlib.exceptions import PluginlibError, PluginImportError, PluginWarning, EntryPointWarning



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Parent Submodule**

Provides plugin bases class and Parent decorator
"""

import inspect
import sys
import warnings

from pluginlib.exceptions import PluginWarning
from pluginlib._objects import GroupDict, TypeDict, PluginDict
from pluginlib._util import (allow_bare_decorator, ClassProperty, DictWithDotNotation,
                             LOGGER, PY26, Undefined)

try:
    from asyncio import iscoroutinefunction
except ImportError:  # pragma: no cover
    iscoroutinefunction = lambda func: False  # noqa: E731


DEFAULT = '_default'
UNDEFINED = Undefined()

isfunction = inspect.isfunction  # pylint: disable=invalid-name
getfullargspec = getattr(inspect,  # pylint: disable=invalid-name
                         'getfullargspec', getattr(inspect, 'getargspec'))

try:
    STR = unicode

except NameError:
    STR = str


class ClassInspector(object):
    """
    Args:
        cls(:py:class:`Plugin`): Parent class
        subclass(:py:class:`Plugin`): Subclass to evaluate

    Inspects subclass for inclusion
    After initialization, the following attributes are set:

    errorcode
        A code corresponding to the error condition

    message
        Error message

    Values for errorcode:

        * 0: No error

        Error codes between 0 and 100 are not intended for import

        * 50 Skipload flag is True

        Error codes between 99 and 200 are excluded from import

        * 156: Skipload call returned True

        Error codes 200 and above are malformed classes

        * 210: Missing abstract property
        * 211: Missing abstract static method
        * 212: Missing abstract class method
        * 213: Missing abstract method
        * 214: Missing abstract attribute
        * 215: Missing abstract coroutine
        * 216: Type annotations differ
        * 220: Argument spec does not match
    """

    def __init__(self, cls, subclass):

        self.message = None
        self.errorcode = 0
        self.cls = cls
        self.subclass = subclass

        self._check_skipload()
        if not self.errorcode:
            self._check_methods()

    def __bool__(self):
        return self.errorcode == 0

    # Python 2 equivalent
    __nonzero__ = __bool__

    def _check_skipload(self):
        """
        Determine if subclass should be skipped
        _skipload_ is either a Boolean or callable that returns a Boolean
        """

        # pylint: disable=protected-access
        if callable(self.subclass._skipload_):

            result = self.subclass._skipload_()

            if isinstance(result, tuple):
                skip, self.message = result
            else:
                skip = result

            if skip:
                self.errorcode = 156

        elif self.subclass._skipload_:
            self.errorcode = 50
            self.message = 'Skipload flag is True'

    def _check_methods(self):
        """
        Validate abstract methods are defined in subclass
        """

        for name, method in self.cls.__abstractmethods__.items():

            if self.errorcode:
                break

            # Need to get attribute from dictionary for instance tests to work
            for base in self.subclass.__mro__:  # pragma: no branch
                if name in base.__dict__:
                    submethod = base.__dict__[name]
                    break

            # If we found our abstract method, we didn't find anything
            if submethod is method:
                submethod = UNDEFINED

            if isinstance(method, property):
                self._check_property(name, submethod)
            elif isinstance(method, staticmethod):
                self._check_static_method(name, method, submethod)
            elif isinstance(method, classmethod):
                self._check_class_method(name, method, submethod)
            elif isfunction(method):
                self._check_generic_method(name, method, submethod)

            # If it's not a type we're specifically checking, just check for existence
            elif submethod is UNDEFINED:
                self.errorcode = 214
                self.message = 'Does not contain required attribute (%s)' % name

            if not self.errorcode:
                self._check_coroutine_method(name, method, submethod)

            if not self.errorcode:
                self._check_annotations(name, method, submethod)

    def _check_property(self, name, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        Check for class properties
        """

        if submethod is UNDEFINED or not isinstance(submethod, property):
            self.errorcode = 210
            self.message = 'Does not contain required property (%s)' % name

    def _check_static_method(self, name, method, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        Check for static methods
        """

        if submethod is UNDEFINED or not isinstance(submethod, staticmethod):
            self.errorcode = 211
            self.message = 'Does not contain required static method (%s)' % name
        elif PY26:  # pragma: no cover
            self._compare_argspec(name, getfullargspec(method.__get__(True)),
                                  getfullargspec(submethod.__get__(True)))
        else:
            self._compare_argspec(name, getfullargspec(method.__func__),
                                  getfullargspec(submethod.__func__))

    def _check_class_method(self, name, method, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        Check for class methods
        """

        if submethod is UNDEFINED or not isinstance(submethod, classmethod):
            self.errorcode = 212
            self.message = 'Does not contain required class method (%s)' % name
        elif PY26:  # pragma: no cover
            self._compare_argspec(name, getfullargspec(method.__get__(True).__func__),
                                  getfullargspec(submethod.__get__(True).__func__))
        else:
            self._compare_argspec(name, getfullargspec(method.__func__),
                                  getfullargspec(submethod.__func__))

    def _check_generic_method(self, name, method, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        Check for generic methods
        """

        if submethod is UNDEFINED or not isfunction(submethod):
            self.errorcode = 213
            self.message = 'Does not contain required method (%s)' % name
        else:
            self._compare_argspec(name, getfullargspec(method), getfullargspec(submethod))

    def _check_coroutine_method(self, name, method, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        If abstract is a coroutine method, child should be too
        """

        if iscoroutinefunction(method) and not iscoroutinefunction(submethod):
            self.errorcode = 215
            self.message = 'Does not contain required coroutine method (%s)' % name

    def _check_annotations(self, name, method, submethod):
        """
        Args:
            name(str): Method name
            method(:py:class:`function`): Abstract method object
            submethod(:py:class:`function`): Subclass method object

        If abstract has type annotations and the child has type annotations, they should match
        """

        meth_annotations = getattr(method, '__annotations__', {})
        if meth_annotations:
            submeth_annotations = getattr(submethod, '__annotations__', {})
            if submeth_annotations and meth_annotations != submeth_annotations:
                self.errorcode = 216
                self.message = 'Type annotations differ for (%s)' % name

    def _compare_argspec(self, name, spec_1, spec_2):
        """
        Args:
            name(str): Method name
            spec_1(:py:class:`inspect.FullArgSpec`): Argument spec
            spec_2(:py:class:`inspect.FullArgSpec`): Argument spec

        Compares two argspecs skipping type annotations
        """

        spec_1_dict = spec_1._asdict()
        spec_2_dict = spec_2._asdict()

        matches = True

        for key, val in spec_1_dict.items():
            # Annotations are checked separately
            if key == 'annotations':
                continue
            if spec_2_dict[key] != val:
                matches = False
                break

        if not matches:
            self.errorcode = 220
            self.message = 'Argument spec does not match parent for method %s' % name


class PluginType(type):
    """
    Metaclass for plugins
    """

    __plugins = DictWithDotNotation([(DEFAULT, GroupDict())])

    # pylint: disable=bad-mcs-classmethod-argument
    def __new__(cls, name, bases, namespace, **kwargs):

        # Make sure '_parent_', '_skipload_' are set explicitly or ignore
        for attr in ('_parent_', '_skipload_'):
            if attr not in namespace:
                namespace[attr] = False

        new = super(PluginType, cls).__new__(cls, name, bases, namespace, **kwargs)

        # Determine group
        group = cls.__plugins.setdefault(new._group_ if new._group_ else DEFAULT, GroupDict())

        if new._type_ in group:
            if new._parent_:
                raise ValueError('parent must be unique: %s' % new._type_)

            plugindict = group[new._type_].get(new.name, UNDEFINED)
            version = STR(new.version or 0)

            # Check for duplicates. Warn and ignore
            if plugindict and version in plugindict:

                existing = plugindict[version]
                warnings.warn("Duplicate plugins found for %s: %s.%s and %s.%s" %
                              (new, new.__module__, new.__name__,
                               existing.__module__, existing.__name__),
                              PluginWarning, stacklevel=2)

            else:
                result = ClassInspector(group[new._type_]._parent, new)

                if result:
                    group[new._type_].setdefault(new.name, PluginDict())[version] = new

                else:
                    skipmsg = u'Skipping %s class %s.%s: Reason: %s'
                    args = (new, new.__module__, new.__name__, result.message)

                    if result.errorcode < 100:
                        LOGGER.debug(skipmsg, *args)
                    elif result.errorcode < 200:
                        LOGGER.info(skipmsg, *args)
                    else:
                        warnings.warn(skipmsg % args, PluginWarning, stacklevel=2)

        elif new._parent_:
            group[new._type_] = TypeDict(new)

            new.__abstractmethods__ = dict()

            # Get abstract methods by walking the MRO
            for base in reversed(new.__mro__):
                for method_name, method in base.__dict__.items():
                    if getattr(method, '__isabstractmethod__', False):
                        new.__abstractmethods__[method_name] = method

        else:
            raise ValueError('Unknown parent type: %s' % new._type_)

        return new

    def _get_plugins(cls):
        """
        Return registered plugins
        """

        return cls.__plugins[cls._group_ if cls._group_ else DEFAULT][cls._type_]


class Plugin(object):
    """
    **Mixin class for plugins.
    All parents and child plugins will inherit from this class automatically.**

    **Class Attributes**

        *The following attributes can be set as class attributes in subclasses*

        .. autoattribute:: _alias_
        .. autoattribute:: _skipload_
        .. autoattribute:: _version_

    **Class Properties**

        .. autoattribute:: name
            :annotation:

            :py:class:`str` -- :attr:`_alias_` if set or falls back to class name

        .. autoattribute:: plugin_group

        .. autoattribute:: plugin_type

        .. autoattribute:: version
            :annotation:

            :py:class:`str` -- Returns :attr:`_version_` if set,
            otherwise falls back to module ``__version__`` or :py:data:`None`

    """

    __slots__ = ()

    _alias_ = None
    """:py:class:`str` -- Friendly name to refer to plugin.
       Accessed through :attr:`~Plugin.name` property."""
    _skipload_ = False
    """:py:class:`bool` -- When True, plugin is not loaded.
       Can also be a static or class method that returns a tuple ``(bool, message)``"""
    _version_ = None
    """:py:class:`str` -- Plugin version. Should adhere to `PEP 440`_.
       Accessed through :attr:`~Plugin.version` property.

       .. _PEP 440: https://www.python.org/dev/peps/pep-0440/"""

    @ClassProperty
    def version(cls):  # noqa: N805  # pylint: disable=no-self-argument
        """
        :py:class:Returns `str` -- Returns :attr:`_version_` if set,
        otherwise falls back to module `__version__` or None
        """
        return cls._version_ or getattr(sys.modules.get(cls.__module__, None),
                                        '__version__', None)

    @ClassProperty
    def name(cls):  # noqa: N805  # pylint: disable=no-self-argument
        """
        :py:class:`str` -- :attr:`_alias_` if set or falls back to class name
        """

        return cls._alias_ or cls.__name__  # pylint: disable=no-member

    @ClassProperty
    def plugin_type(cls):  # noqa: N805  # pylint: disable=no-self-argument
        """
        :py:class:`str` -- ``plugin_type`` of :py:class:`~pluginlib.Parent` class
        """

        return cls._type_  # pylint: disable=no-member

    @ClassProperty
    def plugin_group(cls):  # noqa: N805  # pylint: disable=no-self-argument
        """
        :py:class:`str` -- ``group`` of :py:class:`~pluginlib.Parent` class
        """

        return cls._group_  # pylint: disable=no-member


@allow_bare_decorator
class Parent(object):
    """
    Args:
        plugin_type(str): Plugin type
        group(str): Group to store plugins

    **Class Decorator for plugin parents**

    ``plugin_type`` determines under what attribute child plugins will be accessed in
    :py:attr:`PluginLoader.plugins`.
    When not specified, the class name is used.

    ``group`` specifies the parent and all child plugins are members of
    the specified plugin group. A :py:attr:`PluginLoader` instance only accesses the
    plugins group specified when it was initialized.
    When not specified, the default group is used.
    ``group`` should be specified if plugins for different projects could be accessed
    in an single program, such as in libraries and frameworks.
    """

    __slots__ = 'plugin_type', 'group'

    def __init__(self, plugin_type=None, group=None):
        self.plugin_type = plugin_type
        self.group = group

    def __call__(self, cls):

        # In case we're inheriting another parent, clean registry
        if isinstance(cls, PluginType):
            plugins = cls._get_plugins().get(cls.name, {})
            remove = []

            for pver, pcls in plugins.items():
                if cls is pcls:
                    remove.append(pver)

            if plugins and len(remove) == len(plugins):
                del cls._get_plugins()[cls.name]

            else:
                for key in remove:
                    del plugins[key]

        dict_ = cls.__dict__.copy()
        dict_.pop('__dict__', None)
        dict_.pop('__weakref__', None)

        # Clean out slot members
        for member in dict_.get('__slots__', ()):
            dict_.pop(member, None)

        # Set type
        dict_['_type_'] = self.plugin_type or cls.__name__
        dict_['_group_'] = self.group

        # Mark as parents
        for attr in ('_parent_', '_skipload_'):
            dict_[attr] = True

        # Reorder bases so any overrides are hit first in the MRO
        bases = [base for base in cls.__bases__ if base not in (object, Plugin)]
        bases.append(Plugin)
        bases = tuple(bases)

        return PluginType(cls.__name__, bases, dict_)


def get_plugins():
    """
    Convenience method for accessing all imported plugins
    """

    # pylint: disable=protected-access
    return PluginType._PluginType__plugins



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Exceptions Submodule**

Provides exceptions classes
"""


class PluginlibError(Exception):
    """
    **Base exception class for Pluginlib exceptions**

    All Pluginlib exceptions are derived from this class.

    Subclass of :py:exc:`Exception`

    **Custom Instance Attributes**

        .. py:attribute:: friendly
            :annotation: = None

            :py:class:`str` -- Optional friendly output
    """

    def __init__(self, *args, **kwargs):
        super(PluginlibError, self).__init__(*args)
        self.friendly = kwargs.get('friendly', None)


class PluginImportError(PluginlibError):
    """
    **Exception class for Pluginlib import errors**

    Subclass of :py:exc:`PluginlibError`

    **Custom Instance Attributes**

        .. py:attribute:: friendly
            :annotation: = None

            :py:class:`str` -- May contain abbreviated traceback

            When an exception is raised while importing a module, an attempt is made to create a
            "friendly" version of the output with a traceback limited to the plugin itself
            or, failing that, the loader module.

    """


class PluginWarning(UserWarning):
    """
    Warning for errors with imported plugins

    Subclass of :py:exc:`UserWarning`
    """


class EntryPointWarning(ImportWarning):
    """
    Warning for errors with importing entry points

    Subclass of :py:exc:`ImportWarning`
    """



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Electric engine**
Test data standard library
"""

from __future__ import absolute_import
from tests.testdata.parents import Engine


class Electric(Engine):
    """Dummy electric engine"""

    _alias_ = 'electric'

    def start(self):
        return 'beep'



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Utility Module**

This module contains generic functions for use in other modules
"""

from abc import abstractmethod
from functools import update_wrapper, wraps
from inspect import isclass
import logging
import operator as _operator
import sys


PY26 = sys.version_info[:2] < (2, 7)
PY34 = sys.version_info[:2] < (3, 5)


# Setup logger
if PY26:  # pragma: no branch

    class NullHandler(logging.Handler):  # pragma: no cover
        """NullHandler for Python 2.6"""
        def emit(self, record):
            pass
else:
    NullHandler = logging.NullHandler


LOGGER = logging.getLogger('pluginlib')
LOGGER.addHandler(NullHandler())

OPERATORS = {'=': _operator.eq,
             '==': _operator.eq,
             '!=': _operator.ne,
             '<': _operator.lt,
             '<=': _operator.le,
             '>': _operator.gt,
             '>=': _operator.ge}


NoneType = type(None)

if sys.version_info[0] < 3:  # pragma: no branch
    BASESTRING = basestring  # pragma: no cover  # noqa: F821 # pylint: disable=undefined-variable
else:
    BASESTRING = str


def raise_with_traceback(exc, tback):  # pragma: no cover
    """
    Placeholder for version-specific implementation
    """
    raise NotImplementedError


# Attempt to make a Python 2/3 compatible way to raise with a traceback
# pylint: disable=exec-used
if sys.version_info[0] < 3:  # pragma: no branch
    exec("""def raise_with_traceback(exc, tback):  # pragma: no cover
    raise exc.__class__, exc, tback
""")
else:
    exec("""def raise_with_traceback(exc, tback):
    raise exc.with_traceback(tback) from None
""")


def raise_from_none(exc):  # pragma: no cover
    """
    Convenience function to raise from None in a Python 2/3 compatible manner
    """
    raise exc


if sys.version_info[0] >= 3:  # pragma: no branch
    exec('def raise_from_none(exc):\n    raise exc from None')  # pylint: disable=exec-used


class ClassProperty(object):
    """
    Property decorator for class methods
    """

    def __init__(self, method):

        self.method = method
        update_wrapper(self, method)

    def __get__(self, instance, cls):
        return self.method(cls)


class Undefined(object):
    """
    Class for creating unique undefined objects for value comparisons
    """

    def __init__(self, label='UNDEF'):
        self.label = label

    def __str__(self):
        return self.label

    def __repr__(self):
        return self.__str__()

    def __bool__(self):
        return False

    # For Python 2
    __nonzero__ = __bool__


def allow_bare_decorator(cls):
    """
    Wrapper for a class decorator which allows for bare decorator and argument syntax
    """

    @wraps(cls)
    def wrapper(*args, **kwargs):
        """"Wrapper for real decorator"""

        # If we weren't only passed a bare class, return class instance
        if kwargs or len(args) != 1 or not isclass(args[0]):  # pylint: disable=no-else-return
            return cls(*args, **kwargs)
        # Otherwise, pass call to instance with default values
        else:
            return cls()(args[0])

    return wrapper


class CachingDict(dict):
    """
    A subclass of :py:class:`dict` that has a private _cache attribute

    self._cache is regular dictionary which is cleared whenever the CachingDict is changed

    Nothing is actually cached. That is the responsibility of the inheriting class
    """

    def __init__(self, *args, **kwargs):
        super(CachingDict, self).__init__(*args, **kwargs)
        self._cache = {}

    def __setitem__(self, key, value):
        try:
            super(CachingDict, self).__setitem__(key, value)
        finally:
            self._cache.clear()

    def __delitem__(self, key):
        try:
            super(CachingDict, self).__delitem__(key)
        finally:
            self._cache.clear()

    def clear(self):
        try:
            super(CachingDict, self).clear()
        finally:
            self._cache.clear()

    def setdefault(self, key, default=None):

        try:
            return self[key]
        except KeyError:
            self.__setitem__(key, default)
            return default

    def pop(self, *args):
        try:
            value = super(CachingDict, self).pop(*args)
        except KeyError as e:
            raise e
        else:
            self._cache.clear()
            return value

    def popitem(self):
        try:
            item = super(CachingDict, self).popitem()
        except KeyError as e:
            raise e
        else:
            self._cache.clear()
            return item


class DictWithDotNotation(dict):
    """
    Dictionary addressable by dot notation
    """

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            raise_from_none(AttributeError("'dict' object has no attribute '%s'" % name))


class abstractstaticmethod(staticmethod):  # noqa: N801  # pylint: disable=invalid-name
    """
    A decorator for abstract static methods

    Used in parent classes to identify static methods required in child plugins

    This decorator is included to support older versions of Python and
    should be considered deprecated as of Python 3.3.
    The preferred implementation is:

    .. code-block:: python

        @staticmethod
        @pluginlib.abstractmethod
        def abstract_staticmethod():
            return 'foo'
    """

    __isabstractmethod__ = True

    def __init__(self, func):
        super(abstractstaticmethod, self).__init__(abstractmethod(func))


class abstractclassmethod(classmethod):  # noqa: N801  # pylint: disable=invalid-name
    """
    A decorator for abstract class methods

    Used in parent classes to identify class methods required in child plugins

    This decorator is included to support older versions of Python and
    should be considered deprecated as of Python 3.3.
    The preferred implementation is:

    .. code-block:: python

        @classmethod
        @pluginlib.abstractmethod
        def abstract_classmethod(cls):
            return cls.foo
    """

    __isabstractmethod__ = True

    def __init__(self, func):
        super(abstractclassmethod, self).__init__(abstractmethod(func))


class abstractattribute(object):  # noqa: N801  # pylint: disable=invalid-name
    """
    A class to be used to identify abstract attributes

    .. code-block:: python

        @pluginlib.Parent
        class ParentClass(object):
            abstract_attribute = pluginlib.abstractattribute

    """
    __isabstractmethod__ = True



# Copyright 2017 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Bad Module that will raise syntax error when imported
"""


def func1():
    """Raises an exception"""
    raise RuntimeError('This parrot is no more')


def func2():
    """Calls function 1"""
    func1()


def func3():
    """Calls function 2"""
    func2()


func3()



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**pluginlib test data parents**
Parent classes for tests
"""

from pluginlib import Parent, abstractmethod


@Parent('parser', 'testdata')
class Parser(object):
    """Parser parent class"""

    @abstractmethod
    def parse(self):
        """Required method"""


@Parent('engine', 'testdata')
class Engine(object):
    """Engine parent class"""

    @abstractmethod
    def start(self):
        """Required method"""


@Parent('hook', 'testdata')
class Hook(object):
    """Hook parent class"""



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib._objects**
"""

import pluginlib._objects as objects

from tests import TestCase, mock


# pylint: disable=protected-access


class TestBlacklistEntry(TestCase):
    """Tests for BlacklistEntry class"""

    def test_empty(self):
        """Some value must be supplied or all plugins would be blacklisted"""
        with self.assertRaises(AttributeError):
            objects.BlacklistEntry()
        with self.assertRaises(AttributeError):
            objects.BlacklistEntry(operator='>')

    def test_ver_op(self):
        """Operator and version can be swapped"""
        entry = objects.BlacklistEntry('parser', 'json', '1.0', '>=')
        self.assertEqual(entry.type, 'parser')
        self.assertEqual(entry.name, 'json')
        self.assertEqual(entry.version, '1.0')
        self.assertEqual(entry.operator, '>=')

    def test_op_ver(self):
        """Operator and version can be swapped"""
        entry = objects.BlacklistEntry('parser', 'json', '>=', '1.0')
        self.assertEqual(entry.type, 'parser')
        self.assertEqual(entry.name, 'json')
        self.assertEqual(entry.version, '1.0')
        self.assertEqual(entry.operator, '>=')

    def test_op_none(self):
        """Operator is required when version is supplied"""
        with self.assertRaises(AttributeError):
            objects.BlacklistEntry('parser', 'json', '>=')
        entry = objects.BlacklistEntry('parser', 'json')
        self.assertEqual(entry.type, 'parser')
        self.assertEqual(entry.name, 'json')
        self.assertEqual(entry.version, None)
        self.assertEqual(entry.operator, '==')

    def test_no_op(self):
        """Operator defaults to '=='"""
        entry = objects.BlacklistEntry('parser', 'json', '1.0')
        self.assertEqual(entry.type, 'parser')
        self.assertEqual(entry.name, 'json')
        self.assertEqual(entry.version, '1.0')
        self.assertEqual(entry.operator, '==')

    def test_op_bad(self):
        """Operators must be in '=', '==', '!=', '<', '<=', '>', '>='"""
        with self.assertRaises(AttributeError):
            objects.BlacklistEntry('parser', 'json', '1.0', '%')

    def test_ver_bad(self):
        """Version is not a string"""
        with self.assertRaises(TypeError):
            objects.BlacklistEntry('parser', 'json', 1.0)

    def test_repr(self):
        """Ensure repr displays properly"""
        entry = objects.BlacklistEntry('parser', 'json', '1.0', '>=')
        self.assertEqual(repr(entry), "BlacklistEntry('parser', 'json', '>=', '1.0')")


class TestGroupDict(TestCase):
    """Tests for GroupDict dictionary subclass"""

    def setUp(self):
        """Create sample dictionary"""
        self.mock_type_parser = mock.Mock()
        self.mock_type_parser._filter.return_value = 'parsertype'
        self.mock_type_engine = mock.Mock()
        self.mock_type_engine._filter.return_value = 'enginetype'
        self.mock_type_empty = mock.Mock()
        self.mock_type_empty._filter.return_value = {}

        self.expected = {'parser': 'parsertype', 'engine': 'enginetype', 'empty': {}}
        self.gdict = objects.GroupDict({'parser': self.mock_type_parser,
                                        'engine': self.mock_type_engine,
                                        'empty': self.mock_type_empty})

    def test_no_blacklist(self):
        """Skip all blacklist logic"""
        self.assertEqual(self.gdict._filter(), self.expected)
        self.mock_type_parser._filter.assert_called_with(None, newest_only=False)
        self.mock_type_engine._filter.assert_called_with(None, newest_only=False)
        self.mock_type_empty._filter.assert_called_with(None, newest_only=False)

        self.assertEqual(self.gdict._filter([]), self.expected)
        self.mock_type_parser._filter.assert_called_with(None, newest_only=False)
        self.mock_type_engine._filter.assert_called_with(None, newest_only=False)
        self.mock_type_empty._filter.assert_called_with(None, newest_only=False)

    def test_blacklist_by_type(self):
        """Entire type is blacklisted, so it should be empty"""
        blacklist = [objects.BlacklistEntry('parser')]
        self.expected['parser'] = {}
        self.assertEqual(self.gdict._filter(blacklist), self.expected)
        self.mock_type_parser._filter.assert_not_called()
        self.mock_type_engine._filter.assert_called_with([], newest_only=False)
        self.mock_type_empty._filter.assert_called_with([], newest_only=False)

    def test_blacklist_by_type_name(self):
        """A type and specific name is blacklisted, so blacklist is passed to child for type"""
        blacklist = [objects.BlacklistEntry('parser', 'json')]
        self.assertEqual(self.gdict._filter(blacklist), self.expected)
        self.mock_type_parser._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_type_engine._filter.assert_called_with([], newest_only=False)
        self.mock_type_empty._filter.assert_called_with([], newest_only=False)

    def test_blacklist_by_name(self):
        """A name is blacklisted in all types, so all children are passed the blacklist"""
        blacklist = [objects.BlacklistEntry(None, 'json')]
        self.assertEqual(self.gdict._filter(blacklist), self.expected)
        self.mock_type_parser._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_type_engine._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_type_empty._filter.assert_called_with(blacklist, newest_only=False)

    def test_empty(self):
        """Unless the entire type is blacklisted, an empty return value is still included"""
        blacklist = [objects.BlacklistEntry('empty', 'json')]
        self.assertEqual(self.gdict._filter(blacklist), self.expected)
        self.mock_type_parser._filter.assert_called_with([], newest_only=False)
        self.mock_type_engine._filter.assert_called_with([], newest_only=False)
        self.mock_type_empty._filter.assert_called_with(blacklist, newest_only=False)

    def test_type_filter(self):
        """Types not in list should be ignored"""
        del self.expected['engine']
        self.assertEqual(self.gdict._filter(type_filter=('parser', 'empty')), self.expected)

    def test_type(self):
        """Get a specific type"""
        self.assertEqual(self.gdict._filter(type='engine'), 'enginetype')
        self.mock_type_engine._filter.assert_called_with(None, newest_only=False, type='engine')
        self.assertIsNone(self.gdict._filter(type='penguin'))

    def test_type_and_type_filter(self):
        """Type filter still affects calls for specific types"""
        self.assertEqual(self.gdict._filter(type_filter=('parser', 'empty'), type='engine'), None)
        self.mock_type_engine._filter.assert_not_called()
        self.mock_type_parser._filter.assert_not_called()
        self.mock_type_empty._filter.assert_not_called()

        self.assertEqual(self.gdict._filter(type_filter=('engine', 'empty'), type='engine'),
                         'enginetype')
        self.mock_type_engine._filter.assert_called_with(None, newest_only=False, type='engine')
        self.mock_type_parser._filter.assert_not_called()
        self.mock_type_empty._filter.assert_not_called()


class TestTypeDict(TestCase):
    """Tests for GroupDict dictionary subclass"""

    def setUp(self):
        """Create sample dictionary"""
        self.mock_plugin_json = mock.Mock()
        self.mock_plugin_json._filter.return_value = 'jsonplugin'
        self.mock_plugin_xml = mock.Mock()
        self.mock_plugin_xml._filter.return_value = 'xmlplugin'
        self.expected = {'json': 'jsonplugin', 'xml': 'xmlplugin'}
        self.tdict = objects.TypeDict('parser', {'json': self.mock_plugin_json,
                                                 'xml': self.mock_plugin_xml})

    def test_parent(self):
        """Ensure parent attribute is set"""
        self.assertEqual('parser', self.tdict._parent)

    def test_no_blacklist(self):
        """Skip all blacklist logic"""
        self.assertEqual(self.tdict._filter(), self.expected)
        self.mock_plugin_json._filter.assert_called_with(None, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with(None, newest_only=False)

        self.assertEqual(self.tdict._filter([]), self.expected)
        self.mock_plugin_json._filter.assert_called_with(None, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with(None, newest_only=False)

    def test_blacklist_by_name(self):
        """Entire name is blacklisted, so it's not called or included"""
        blacklist = [objects.BlacklistEntry(None, 'json')]
        del self.expected['json']
        self.assertEqual(self.tdict._filter(blacklist), self.expected)
        self.mock_plugin_json._filter.assert_not_called()
        self.mock_plugin_xml._filter.assert_called_with([], newest_only=False)

    def test_blacklist_all(self):
        """Type is blacklisted, so all are blacklisted"""
        blacklist = [objects.BlacklistEntry('parser')]
        self.assertEqual(self.tdict._filter(blacklist), {})
        self.mock_plugin_json._filter.assert_not_called()
        self.mock_plugin_xml._filter.assert_not_called()

    def test_blacklist_by_name_version(self):
        """A name and version is blacklisted, so blacklist is passed to child for type"""
        blacklist = [objects.BlacklistEntry('parser', 'json', '1.0')]
        self.assertEqual(self.tdict._filter(blacklist), self.expected)
        self.mock_plugin_json._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with([], newest_only=False)

    def test_blacklist_by_version(self):
        """Only a version is blacklisted, so blacklist is passed to all children"""
        blacklist = [objects.BlacklistEntry(None, None, '1.0')]
        self.assertEqual(self.tdict._filter(blacklist), self.expected)
        self.mock_plugin_json._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with(blacklist, newest_only=False)

    def test_empty(self):
        """Empty values are not included"""
        mock_plugin_empty = mock.Mock()
        mock_plugin_empty._filter.return_value = None
        self.tdict['empty'] = mock_plugin_empty
        self.assertEqual(self.tdict._filter(), self.expected)
        self.tdict._filter()
        self.mock_plugin_json._filter.assert_called_with(None, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with(None, newest_only=False)
        mock_plugin_empty._filter.assert_called_with(None, newest_only=False)

        blacklist = [objects.BlacklistEntry(None, None, '1.0')]
        self.assertEqual(self.tdict._filter(blacklist), self.expected)
        self.mock_plugin_json._filter.assert_called_with(blacklist, newest_only=False)
        self.mock_plugin_xml._filter.assert_called_with(blacklist, newest_only=False)
        mock_plugin_empty._filter.assert_called_with(blacklist, newest_only=False)

    def test_type_filter(self):
        """type_filter ignored in type dict"""
        self.assertEqual(self.tdict._filter(type_filter=('engine')), self.expected)

    def test_name(self):
        """Get a specific name"""
        self.assertEqual(self.tdict._filter(name='json'), 'jsonplugin')
        self.mock_plugin_json._filter.assert_called_with(None, newest_only=False, name='json')

        self.assertIsNone(self.tdict._filter(name='pengion'))


class TestPluginDict(TestCase):
    """Tests for PluginDict dictionary subclass"""

    def setUp(self):
        self.udict = objects.PluginDict({'2.0': 'dos', '1.0': 'uno', '3.0': 'tres'})

    def test_sorted_keys(self):
        """Sorts by version and populates cache"""

        sorted_keys = ['1.0', '2.0', '3.0']
        self.assertFalse('sorted_keys' in self.udict._cache)

        self.assertEqual(self.udict._sorted_keys(), sorted_keys)
        self.assertTrue('sorted_keys' in self.udict._cache)
        self.assertEqual(self.udict._cache['sorted_keys'], sorted_keys)

        # Call again to make sure results are consistent
        self.assertEqual(self.udict._sorted_keys(), sorted_keys)
        self.assertTrue('sorted_keys' in self.udict._cache)
        self.assertEqual(self.udict._cache['sorted_keys'], sorted_keys)

        del self.udict['2.0']
        self.assertFalse('sorted_keys' in self.udict._cache)
        self.assertEqual(self.udict._sorted_keys(), ['1.0', '3.0'])
        self.assertTrue('sorted_keys' in self.udict._cache)
        self.assertEqual(self.udict._cache['sorted_keys'], ['1.0', '3.0'])

    def test_empty(self):
        """Empty dictionary will return None"""
        udict = objects.PluginDict()
        self.assertIsNone(udict._filter())
        self.assertIsNone(udict._filter(newest_only=True))
        self.assertIsNone(udict._filter('FakeBlackList'))
        self.assertIsNone(udict._filter('FakeBlackList', newest_only=True))

    def test_no_blacklist(self):
        """Return newest without filtering"""
        self.assertEqual(self.udict._filter(newest_only=True), 'tres')

    def test_empty_blacklist(self):
        """Same as no blacklist"""
        self.assertEqual(self.udict._filter([]), self.udict)
        self.assertEqual(self.udict._filter([], newest_only=True), 'tres')

    def test_blacklist(self):
        """Cache gets populated and results get filtered"""

        # Cache is populated and highest value is filtered
        blacklist = [objects.BlacklistEntry(None, None, '3.0')]
        self.assertFalse('blacklist' in self.udict._cache)
        self.assertEqual(self.udict._filter(blacklist, newest_only=True), 'dos')
        self.assertEqual(len(self.udict._cache['blacklist']), 1)
        self.assertEqual(self.udict._cache['blacklist'][('3.0', '==')], set(['3.0']))

        self.assertEqual(self.udict._filter(blacklist),
                         objects.OrderedDict([('1.0', 'uno'), ('2.0', 'dos')]))

        # Another entry is cached, result is the same
        blacklist.append(objects.BlacklistEntry(None, None, '1.0'))
        self.assertEqual(self.udict._filter(blacklist, newest_only=True), 'dos')
        self.assertEqual(len(self.udict._cache['blacklist']), 2)
        self.assertEqual(self.udict._cache['blacklist'][('1.0', '==')], set(['1.0']))

        self.assertEqual(self.udict._filter(blacklist),
                         objects.OrderedDict([('2.0', 'dos')]))

        # An equivalent entry is added, cache is the same
        blacklist.append(objects.BlacklistEntry(None, 'number', '3.0', '=='))
        self.assertEqual(self.udict._filter(blacklist, newest_only=True), 'dos')
        self.assertEqual(len(self.udict._cache['blacklist']), 2)

        self.assertEqual(self.udict._filter(blacklist),
                         objects.OrderedDict([('2.0', 'dos')]))

    def test_no_result(self):
        """Blacklist filters all"""
        blacklist = [objects.BlacklistEntry(None, None, '4.0', '<')]

        self.assertEqual(self.udict._filter(blacklist), objects.OrderedDict())
        self.assertEqual(self.udict._cache['blacklist'][('4.0', '<')], set(['1.0', '2.0', '3.0']))

        self.assertIsNone(self.udict._filter(blacklist, newest_only=True))
        self.assertEqual(self.udict._cache['blacklist'][('4.0', '<')], set(['1.0', '2.0', '3.0']))

    def test_version(self):
        """Return specific version"""
        self.assertEqual(self.udict._filter(newest_only=True, version='1.0'), 'uno')

    def test_version_blacklist(self):
        """Return specific version if not blacklisted"""
        blacklist = [objects.BlacklistEntry(None, None, '1.0')]
        self.assertEqual(self.udict._filter(blacklist, newest_only=True, version='1.0'), None)
        self.assertEqual(self.udict._filter(blacklist, newest_only=True, version='2.0'), 'dos')



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Loader Submodule**

Provides functions and classes for loading plugins
"""

import importlib
from inspect import ismodule
import os
import pkgutil
import sys
import traceback
import warnings

from pkg_resources import iter_entry_points, EntryPoint

from pluginlib.exceptions import PluginImportError, EntryPointWarning
from pluginlib._objects import BlacklistEntry
from pluginlib._parent import get_plugins
from pluginlib._util import BASESTRING, LOGGER, NoneType, PY34, raise_with_traceback

try:
    from collections.abc import Iterable
except ImportError:   # pragma: no cover
    # For Python < 3.3
    from collections import Iterable


def format_exception(etype, value, tback, limit=None):
    """
    Python 2 compatible version of traceback.format_exception
    Accepts negative limits like the Python 3 version
    """

    rtn = ['Traceback (most recent call last):\n']

    if limit is None or limit >= 0:
        rtn.extend(traceback.format_tb(tback, limit))
    else:
        rtn.extend(traceback.format_list(traceback.extract_tb(tback)[limit:]))

    rtn.extend(traceback.format_exception_only(etype, value))

    return rtn


def _raise_friendly_exception(exc, name, path):
    """
    Attempt to create a friendly traceback that only shows the errors
    encountered from the plugin import and no the framework
    """

    etype = exc.__class__
    tback = getattr(exc, '__traceback__', sys.exc_info()[2])

    # Create traceback starting at module for friendly output
    start = 0
    here = 0
    tb_list = traceback.extract_tb(tback)

    if path:
        for idx, entry in enumerate(tb_list):
            # Find index for traceback starting with module we tried to load
            if os.path.dirname(entry[0]) == path:
                start = idx
                break

            # Find index for traceback starting with this file
            if os.path.splitext(entry[0])[0] == os.path.splitext(__file__)[0]:
                here = idx

    if start == 0 and isinstance(exc, SyntaxError):
        limit = 0
    else:
        limit = 0 - len(tb_list) + max(start, here)

    # pylint: disable=wrong-spelling-in-comment
    # friendly = ''.join(traceback.format_exception(etype, exc, tback, limit))
    friendly = ''.join(format_exception(etype, exc, tback, limit))

    # Format exception
    msg = 'Error while importing candidate plugin module %s from %s' % (name, path)
    exception = PluginImportError('%s: %s' % (msg, repr(exc)), friendly=friendly)

    raise_with_traceback(exception, tback)


def _import_module(name, path=None):
    """
    Args:
        name(str):
            * Full name of object
            * name can also be an EntryPoint object, name and path will be determined dynamically
        path(str): Module directory

    Returns:
        object: module object or advertised object for EntryPoint

    Loads a module using importlib catching exceptions
    If path is given, the traceback will be formatted to give more friendly and direct information
    """

    # If name is an entry point, try to parse it
    epoint = None
    if isinstance(name, EntryPoint):
        epoint = name
        name = epoint.module_name

    if path is None:
        try:
            loader = pkgutil.get_loader(name)
        except ImportError:
            pass
        else:
            if loader:
                path = os.path.dirname(loader.get_filename(name))

    LOGGER.debug('Attempting to load module %s from %s', name, path)
    try:
        mod = epoint.load() if epoint else importlib.import_module(name)

    except Exception as e:  # pylint: disable=broad-except
        _raise_friendly_exception(e, name, path)

    return mod


def _recursive_import(package):
    """
    Args:
        package(py:term:`package`): Package to walk

    Import all modules from a package recursively
    """

    prefix = '%s.' % (package.__name__)

    path = getattr(package, '__path__', None)

    if path:
        # pylint: disable=unused-variable
        for finder, name, is_pkg in pkgutil.walk_packages(path, prefix=prefix):
            _import_module(name, finder.path)


def _recursive_path_import(path, prefix_package):
    """
    Args:
        path(str): Path to walk
        prefix_package(str): Prefix to apply to found modules

    Import all modules from a path recursively

    If a python package is found, it is imported recursively,
    however, the directory walk will stop on the directory of the package
    and Python files under the package that are in directories without
    init files, will be skipped.
    """

    # Include basename of path in module prefix
    basename = os.path.basename(path.strip('/'))
    root_prefix = '%s.%s.' % (prefix_package, basename)
    prefix_template = '%s%%s.' % root_prefix

    # Walk path
    for root, dirs, files in os.walk(path):
        # If root is a Python module, we won't walk any farther down it
        # find_module() seems to be recursive in Python 3
        if '__init__.py' in files and (3, 0) <= sys.version_info[:2] < (3, 5):  # pragma: no cover
            dirs.clear()
            if root != path:
                continue

        # Generate prefix
        if root == path:
            prefix = root_prefix
        else:
            relpath = os.path.relpath(root, path)
            prefix = prefix_template % relpath.replace(os.sep, '.')

        # Walk root and import modules
        # pylint: disable=unused-variable
        for finder, name, is_pkg in pkgutil.walk_packages([root], prefix=prefix):
            LOGGER.debug('Attempting to load module %s from %s', name, finder.path)
            try:
                # find_module() was deprecated in 3.4, but module_from_spec wasn't available
                if PY34:  # pragma: no cover
                    finder.find_module(name).load_module(name)
                else:
                    spec = finder.find_spec(name)
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)

            except Exception as e:  # pylint: disable=broad-except
                _raise_friendly_exception(e, name, root)


# pylint: disable=too-many-instance-attributes,too-many-arguments
class PluginLoader(object):
    """
    Args:
        group(str): Group to retrieve plugins from
        library(str): Standard library package
        modules(list): Iterable of modules to import recursively
        paths(list): Iterable of paths to import recursively
        entry_point(str): `Entry point`_ for additional plugins
        blacklist(list): Iterable of :py:class:`BlacklistEntry` objects or tuples
        prefix_package(str): Alternative prefix for imported packages
        type_filter(list): Iterable of parent plugin types to allow

    **Interface for importing and accessing plugins**

    Plugins are loaded from sources specified at initialization when
    :py:meth:`load_modules` is called or when the :py:attr:`plugins` property is first accessed.

    ``group`` specifies the group whose members will be returned by :py:attr:`plugins`
    This corresponds directly with the ``group`` attribute for :py:func:`Parent`.
    When not specified, the default group is used.
    ``group`` should be specified if plugins for different projects could be accessed
    in an single program, such as in libraries and frameworks.

    ``library`` indicates the package of a program's standard library.
    This should be a package which is always loaded.

    ``modules`` is an iterable of optional modules to load.
    If a package is given, it will be loaded recursively.

    ``paths`` is an iterable of optional paths to find modules to load.
    The paths are searched recursively and imported under the namespace specified by
    ``prefix_package``.

    ``entry_point`` specifies an `entry point <Entry point_>`_ group to identify
    additional modules and packages which should be loaded.

    ``blacklist`` is an iterable containing :py:class:`BlacklistEntry` objects or tuples
    with arguments for new :py:class:`BlacklistEntry` objects.

    ``prefix_package`` must be the name of an existing package under which to import the paths
    specified in ``paths``. Because the package paths will be traversed recursively, this should
    be an empty path.

    ``type_filter`` limits plugins types to only those specified. A specified type is not
    guaranteed to be available.

    .. _Entry point: https://packaging.python.org/specifications/entry-points/
    """

    def __init__(self, group=None, library=None, modules=None, paths=None, entry_point=None,
                 blacklist=None, prefix_package='pluginlib.importer', type_filter=None):

        # Make sure we got iterables
        for argname, arg in (('modules', modules), ('paths', paths), ('blacklist', blacklist),
                             ('type_filter', type_filter)):
            if not isinstance(arg, (NoneType, Iterable)) or isinstance(arg, BASESTRING):
                raise TypeError("Expecting iterable for '%s', recieved %s" % (argname, type(arg)))

        # Make sure we got strings
        for argname, arg in (('library', library), ('entry_point', entry_point),
                             ('prefix_package', prefix_package)):
            if not isinstance(arg, (NoneType, BASESTRING)):
                raise TypeError("Expecting string for '%s', recieved %s" % (argname, type(arg)))

        self.group = group or '_default'
        self.library = library
        self.modules = modules or tuple()
        self.paths = paths or tuple()
        self.entry_point = entry_point
        self.prefix_package = prefix_package
        self.type_filter = type_filter
        self.loaded = False

        if blacklist:
            self.blacklist = []
            for entry in blacklist:

                if isinstance(entry, BlacklistEntry):
                    pass
                elif isinstance(entry, Iterable):
                    try:
                        entry = BlacklistEntry(*entry)
                    except (AttributeError, TypeError) as e:
                        # pylint: disable=raise-missing-from
                        raise AttributeError("Invalid blacklist entry '%s': %s " % (entry, e))
                else:
                    raise AttributeError("Invalid blacklist entry '%s': Not an iterable" % entry)

                self.blacklist.append(entry)

            self.blacklist = tuple(self.blacklist)

        else:
            self.blacklist = None

    def __repr__(self):

        args = []
        for attr, default in (('group', '_default'), ('library', None), ('modules', None),
                              ('paths', None), ('entry_point', None), ('blacklist', None),
                              ('prefix_package', 'pluginlib.importer'), ('type_filter', None)):

            val = getattr(self, attr)
            if default and val == default:
                continue

            if val:
                args.append('%s=%r' % (attr, val))

        return '%s(%s)' % (self.__class__.__name__, ', '.join(args))

    def load_modules(self):
        """
        Locate and import modules from locations specified during initialization.

        Locations include:
            - Program's standard library (``library``)
            - `Entry points <Entry point_>`_ (``entry_point``)
            - Specified modules (``modules``)
            - Specified paths (``paths``)

        If a malformed child plugin class is imported, a :py:exc:`PluginWarning` will be issued,
        the class is skipped, and loading operations continue.

        If an invalid `entry point <Entry point_>`_ is specified, an :py:exc:`EntryPointWarning`
        is issued and loading operations continue.
        """

        # Start with standard library
        if self.library:
            LOGGER.info('Loading plugins from standard library')
            libmod = _import_module(self.library)
            _recursive_import(libmod)

        # Get entry points
        if self.entry_point:
            LOGGER.info('Loading plugins from entry points group %s', self.entry_point)
            for epoint in iter_entry_points(group=self.entry_point):
                try:
                    mod = _import_module(epoint)
                except PluginImportError as e:
                    warnings.warn("Module %s can not be loaded for entry point %s: %s" %
                                  (epoint.module_name, epoint.name, e), EntryPointWarning)
                    continue

                # If we have a package, walk it
                if ismodule(mod):
                    _recursive_import(mod)
                else:
                    warnings.warn("Entry point '%s' is not a module or package" % epoint.name,
                                  EntryPointWarning)

        # Load auxiliary modules
        if self.modules:
            for mod in self.modules:
                LOGGER.info('Loading plugins from %s', mod)
                _recursive_import(_import_module(mod))

        # Load auxiliary paths
        if self.paths:
            # Import each path recursively
            for path in self.paths:
                modpath = os.path.realpath(path)
                if os.path.isdir(modpath):
                    LOGGER.info("Recursively importing plugins from path `%s`", path)
                    _recursive_path_import(path, self.prefix_package)
                else:
                    LOGGER.info("Configured plugin path '%s' is not a valid directory", path)

        self.loaded = True

    @property
    def plugins(self):
        """
        Newest version of all plugins in the group filtered by ``blacklist``

        Returns:
            dict: Nested dictionary of plugins accessible through dot-notation.

        Plugins are returned in a nested dictionary, but can also be accessed through dot-notion.
        Just as when accessing an undefined dictionary key with index-notation,
        a :py:exc:`KeyError` will be raised if the plugin type or plugin does not exist.

        Parent types are always included.
        Child plugins will only be included if a valid, non-blacklisted plugin is available.
        """

        if not self.loaded:
            self.load_modules()

        # pylint: disable=protected-access
        return get_plugins()[self.group]._filter(blacklist=self.blacklist, newest_only=True,
                                                 type_filter=self.type_filter)

    @property
    def plugins_all(self):
        """
        All resulting versions of all plugins in the group filtered by ``blacklist``

        Returns:
            dict: Nested dictionary of plugins accessible through dot-notation.

        Similar to :py:attr:`plugins`, but lowest level is an :py:class:`~collections.OrderedDict`
        of all unfiltered plugin versions for the given plugin type and name.

        Parent types are always included.
        Child plugins will only be included if at least one valid, non-blacklisted plugin
        is available.

        The newest plugin can be retrieved by accessing the last item in the dictionary.

        .. code-block:: python

            plugins = loader.plugins_all
            tuple(plugins.parser.json.values())[-1]
        """

        if not self.loaded:
            self.load_modules()

        # pylint: disable=protected-access
        return get_plugins()[self.group]._filter(blacklist=self.blacklist,
                                                 type_filter=self.type_filter)

    def get_plugin(self, plugin_type, name, version=None):
        """
        Args:
            plugin_type(str): Parent type
            name(str): Plugin name
            version(str): Plugin version

        Returns:
            :py:class:`Plugin`: Plugin, or :py:data:`None` if plugin can't be found

        Retrieve a specific plugin. ``blacklist`` and ``type_filter`` still apply.

        If ``version`` is not specified, the newest available version is returned.
        """

        if not self.loaded:
            self.load_modules()

        # pylint: disable=protected-access
        return get_plugins()[self.group]._filter(blacklist=self.blacklist,
                                                 newest_only=True,
                                                 type_filter=self.type_filter,
                                                 type=plugin_type,
                                                 name=name,
                                                 version=version)



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib._loader**
"""

import os
import sys
import warnings

from pkg_resources import Distribution, EntryPoint, working_set

import pluginlib._loader as loader
from pluginlib._objects import OrderedDict
from pluginlib import BlacklistEntry, PluginImportError, EntryPointWarning, PluginWarning

from tests import TestCase, OUTPUT, mock
import tests.testdata
import tests.testdata.parents

try:
    from importlib import reload
except ImportError:
    pass


DATAPATH = os.path.dirname(tests.testdata.__file__)
DIST = Distribution.from_filename(tests.testdata.__file__)
working_set.add(DIST)
DIST._ep_map = {}  # pylint: disable=protected-access


class TestPluginLoaderInit(TestCase):
    """Tests for initialization of PluginLoader"""

    def test_bad_arguments(self):
        """Error is raised when argument type is wrong"""

        # Expect iterables
        for arg in ('modules', 'paths', 'blacklist', 'type_filter'):

            with self.assertRaises(TypeError):
                loader.PluginLoader(**{arg: 'string'})

            with self.assertRaises(TypeError):
                loader.PluginLoader(**{arg: 8675309})

        # Expect strings
        for arg in ('library', 'entry_point', 'prefix_package'):

            with self.assertRaises(TypeError):
                loader.PluginLoader(**{arg: [1, 2, 3]})

            with self.assertRaises(TypeError):
                loader.PluginLoader(**{arg: 8675309})

    def test_blacklist_argument(self):
        """blacklist argument gets verified and processed in init"""

        # No blacklist
        ploader = loader.PluginLoader()
        self.assertIsNone(ploader.blacklist)

        # List of tuples and BlacklistEntry objects
        blentry = BlacklistEntry('parser', 'yaml', '2.0', '>=')
        blacklist = [blentry, ('parser', 'json'), ('parser', 'xml', '<', '1.0')]
        ploader = loader.PluginLoader(blacklist=blacklist)

        self.assertEqual(len(ploader.blacklist), 3)
        for entry in ploader.blacklist:
            self.assertIsInstance(entry, BlacklistEntry)

        self.assertIs(ploader.blacklist[0], blentry)

        self.assertEqual(ploader.blacklist[1].type, 'parser')
        self.assertEqual(ploader.blacklist[1].name, 'json')
        self.assertEqual(ploader.blacklist[1].version, None)
        self.assertEqual(ploader.blacklist[1].operator, '==')

        self.assertEqual(ploader.blacklist[2].type, 'parser')
        self.assertEqual(ploader.blacklist[2].name, 'xml')
        self.assertEqual(ploader.blacklist[2].version, '1.0')
        self.assertEqual(ploader.blacklist[2].operator, '<')

        # Entry isn't iterable
        with self.assertRaisesRegex(AttributeError, 'Invalid blacklist entry'):
            ploader = loader.PluginLoader(blacklist=[1, 2, 3])

        # Bad arguments for BlacklistEntry
        with self.assertRaisesRegex(AttributeError, 'Invalid blacklist entry'):
            ploader = loader.PluginLoader(blacklist=[(1, 2, 3)])

    def test_repr(self):
        """Only non-default settings show up in repr"""

        ploader = loader.PluginLoader(modules=['avengers.iwar', 'avengers.stark'], group='Avengers')
        output = "PluginLoader(group='Avengers', modules=['avengers.iwar', 'avengers.stark'])"
        self.assertEqual(repr(ploader), output)

        blacklist = [('parser', 'json'), ('parser', 'xml', '<', '1.0')]
        output = "PluginLoader(blacklist=(%r, %r))" % (BlacklistEntry('parser', 'json'),
                                                       BlacklistEntry('parser', 'xml', '<', '1.0'))
        ploader = loader.PluginLoader(blacklist=blacklist)
        self.assertEqual(repr(ploader), output)


def unload(module):
    """Unload a package/module and all submodules from sys.modules"""
    for mod in [mod for mod in sys.modules if mod.startswith(module)]:
        del sys.modules[mod]


# pylint: disable=protected-access,too-many-public-methods
class TestPluginLoader(TestCase):
    """Tests for PluginLoader"""

    def setUp(self):

        # Truncate log output
        OUTPUT.seek(0)
        OUTPUT.truncate(0)

    def tearDown(self):

        # Clear entry points
        DIST._ep_map.clear()

        loader.get_plugins().clear()
        unload('tests.testdata.lib')
        unload('pluginlib.importer.')
        reload(tests.testdata.parents)

    def test_load_lib(self):
        """Load modules from standard library"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.parser), 2)
        self.assertTrue('xml' in plugins.parser)
        self.assertTrue('json' in plugins.parser)
        self.assertEqual(plugins.parser.json.version, '2.0')

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('steam' in plugins.engine)
        self.assertEqual(len(plugins.hook), 2)
        self.assertTrue('right' in plugins.hook)
        self.assertTrue('left' in plugins.hook)

    def test_load_entry_points_pkg(self):
        """Load modules from entry points"""

        # Entry point is package
        epoint1 = EntryPoint.parse('hooks = tests.testdata.lib.hooks', dist=DIST)
        # Entry point is module
        epoint2 = EntryPoint.parse('parsers = tests.testdata.lib.parsers.xml', dist=DIST)

        DIST._ep_map = {'pluginlib.test.plugins': {'hooks': epoint1, 'parsers': epoint2}}

        ploader = loader.PluginLoader(group='testdata', entry_point='pluginlib.test.plugins')
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.engine), 0)

        self.assertEqual(len(plugins.hook), 2)
        self.assertTrue('right' in plugins.hook)
        self.assertTrue('left' in plugins.hook)

        self.assertEqual(len(plugins.parser), 1)
        self.assertTrue('xml' in plugins.parser)

    def test_load_entry_points_bad(self):
        """Raise warning and continue when entry point fails - bad package"""

        epoint1 = EntryPoint.parse('bad = not.a.real.module', dist=DIST)
        epoint2 = EntryPoint.parse('parsers = tests.testdata.lib.parsers.xml', dist=DIST)
        DIST._ep_map = {'pluginlib.test.plugins': {'bad': epoint1, 'parsers': epoint2}}

        ploader = loader.PluginLoader(group='testdata', entry_point='pluginlib.test.plugins')

        with warnings.catch_warnings(record=True) as e:
            warnings.simplefilter("always")
            plugins = ploader.plugins

            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, EntryPointWarning))
            self.assertRegex(str(e[-1].message), 'can not be loaded for entry point bad')

        self.assertEqual(len(plugins.parser), 1)
        self.assertTrue('xml' in plugins.parser)
        self.assertEqual(len(plugins.engine), 0)
        self.assertEqual(len(plugins.hook), 0)

    def test_load_entry_points_bad2(self):
        """Raise warning and continue when entry point fails - bad module"""

        epoint1 = EntryPoint.parse('bad = tests.testdata.lib.parsers.bad', dist=DIST)
        epoint2 = EntryPoint.parse('parsers = tests.testdata.lib.parsers.xml', dist=DIST)
        DIST._ep_map = {'pluginlib.test.plugins': {'bad': epoint1, 'parsers': epoint2}}

        ploader = loader.PluginLoader(group='testdata', entry_point='pluginlib.test.plugins')

        with warnings.catch_warnings(record=True) as e:
            warnings.simplefilter("always")
            plugins = ploader.plugins

            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, EntryPointWarning))
            self.assertRegex(str(e[-1].message), 'can not be loaded for entry point bad')

        self.assertEqual(len(plugins.parser), 1)
        self.assertTrue('xml' in plugins.parser)
        self.assertEqual(len(plugins.engine), 0)
        self.assertEqual(len(plugins.hook), 0)

    def test_load_entry_points_not_mod(self):
        """Raise warning and continue when entry point fails"""

        epoint = EntryPoint.parse('parsers = tests.testdata.lib.parsers.xml:XML', dist=DIST)
        DIST._ep_map = {'pluginlib.test.plugins': {'parsers': epoint}}

        ploader = loader.PluginLoader(group='testdata', entry_point='pluginlib.test.plugins')

        with warnings.catch_warnings(record=True) as e:
            warnings.simplefilter("always")
            plugins = ploader.plugins

            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, EntryPointWarning))
            self.assertRegex(str(e[-1].message), "Entry point 'parsers' is not a module or package")

        self.assertEqual(len(plugins.parser), 1)
        self.assertTrue('xml' in plugins.parser)
        self.assertEqual(len(plugins.engine), 0)
        self.assertEqual(len(plugins.hook), 0)

    def test_load_modules(self):
        """Load modules from modules"""
        ploader = loader.PluginLoader(group='testdata', modules=['tests.testdata.lib.parsers'])
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.engine), 0)
        self.assertEqual(len(plugins.hook), 0)

        self.assertEqual(len(plugins.parser), 2)
        self.assertTrue('xml' in plugins.parser)
        self.assertTrue('json' in plugins.parser)
        self.assertEqual(plugins.parser.json.version, '2.0')

    def test_load_paths(self):
        """Load modules from paths"""

        path = os.path.join(DATAPATH, 'lib', 'engines')
        ploader = loader.PluginLoader(group='testdata', paths=[path])
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.parser), 0)
        self.assertEqual(len(plugins.hook), 0)

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('steam' in plugins.engine)

        self.assertEqual(plugins.engine.steam.__module__, 'pluginlib.importer.engines.steam')

    def test_load_paths_bare(self):
        """Load modules from paths without init file"""

        path = os.path.join(DATAPATH, 'bare')
        ploader = loader.PluginLoader(group='testdata', paths=[path])
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.parser), 1)
        self.assertTrue('sillywalk' in plugins.parser)

        self.assertEqual(len(plugins.hook), 2)
        self.assertTrue('fish' in plugins.hook)
        self.assertTrue('grappling' in plugins.hook)

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('electric' in plugins.engine)

        self.assertEqual(plugins.engine.electric.__module__,
                         'pluginlib.importer.bare.engines.electric')

    def test_load_paths_prefix_pkg(self):
        """Load modules from paths with alternate prefix package"""

        path = os.path.join(DATAPATH, 'lib', 'engines')
        ploader = loader.PluginLoader(group='testdata', paths=[path],
                                      prefix_package='tests.testdata.importer')
        plugins = ploader.plugins

        self.assertEqual(len(plugins), 3)
        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertTrue('hook' in plugins)

        self.assertEqual(len(plugins.parser), 0)
        self.assertEqual(len(plugins.hook), 0)

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('steam' in plugins.engine)

        self.assertEqual(plugins.engine.steam.__module__, 'tests.testdata.importer.engines.steam')

    def test_type_filter(self):
        """Filter plugin types"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib',
                                      type_filter=('engine', 'parser'))
        plugins = ploader.plugins

        self.assertTrue('parser' in plugins)
        self.assertTrue('engine' in plugins)
        self.assertFalse('hook' in plugins)

    def test_load_paths_missing(self):
        """Log on invalid path"""

        path1 = os.path.join(DATAPATH, 'lib', 'engines')
        path2 = os.path.join(DATAPATH, 'NotARealPath')
        ploader = loader.PluginLoader(group='testdata', paths=[path1, path2])
        plugins = ploader.plugins

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('steam' in plugins.engine)

        self.assertRegex(OUTPUT.getvalue().splitlines()[-1], 'not a valid directory')

    def test_load_paths_duplicate(self):
        """Ignore duplicate paths"""

        path = os.path.join(DATAPATH, 'lib', 'engines')

        with warnings.catch_warnings(record=True) as e:

            warnings.simplefilter("always")
            ploader = loader.PluginLoader(group='testdata', paths=[path, path])
            plugins = ploader.plugins

            self.assertEqual(len(e), 1)
            self.assertTrue(issubclass(e[-1].category, PluginWarning))
            self.assertRegex(str(e[-1].message), 'Duplicate plugins found')

        self.assertEqual(len(plugins.engine), 1)
        self.assertTrue('steam' in plugins.engine)

    def test_bad_import(self):
        """Syntax error in imported module"""

        ploader = loader.PluginLoader(group='testdata', modules=['tests.testdata.bad.syntax'])
        error = 'Error while importing candidate plugin module tests.testdata.bad.syntax'
        with self.assertRaisesRegex(PluginImportError, error) as e:
            ploader.plugins  # pylint: disable=pointless-statement

        self.assertRegex(e.exception.friendly, 'SyntaxError: invalid syntax')
        self.assertRegex(e.exception.friendly, 'tests.testdata.bad.syntax')
        self.assertRegex(e.exception.friendly, 'line 12')

    def test_bad_import2(self):
        """Exception raised by imported module"""

        ploader = loader.PluginLoader(group='testdata', modules=['tests.testdata.bad2'])
        error = 'Error while importing candidate plugin module tests.testdata.bad2'
        with self.assertRaisesRegex(PluginImportError, error) as e:
            ploader.plugins  # pylint: disable=pointless-statement

        self.assertRegex(e.exception.friendly, 'RuntimeError: This parrot is no more')
        self.assertRegex(e.exception.friendly, 'tests.testdata.bad2')
        self.assertRegex(e.exception.friendly, 'line 24')

    def test_bad_import_path(self):
        """Syntax error in imported module loaded by path"""

        path = os.path.join(DATAPATH, 'bad')
        ploader = loader.PluginLoader(group='testdata', paths=[path])
        error = 'Error while importing candidate plugin module pluginlib.importer.bad.syntax'
        with self.assertRaisesRegex(PluginImportError, error) as e:
            ploader.plugins  # pylint: disable=pointless-statement

        self.assertRegex(e.exception.friendly, 'SyntaxError: invalid syntax')
        self.assertRegex(e.exception.friendly, 'testdata/bad/syntax.py')
        self.assertRegex(e.exception.friendly, 'line 12')

    def test_plugins(self):
        """plugins only loads modules on the first call"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        with mock.patch.object(ploader, 'load_modules',
                               wraps=ploader.load_modules) as mock_load_modules:
            plugins1 = ploader.plugins
            self.assertEqual(mock_load_modules.call_count, 1)

            self.assertEqual(len(plugins1.parser), 2)
            self.assertTrue('xml' in plugins1.parser)
            self.assertTrue('json' in plugins1.parser)
            self.assertEqual(plugins1.parser.json.version, '2.0')

            plugins2 = ploader.plugins
            self.assertEqual(mock_load_modules.call_count, 1)

            self.assertEqual(plugins1, plugins2)

    def test_plugins_all(self):
        """plugins only loads modules on the first call"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        with mock.patch.object(ploader, 'load_modules',
                               wraps=ploader.load_modules) as mock_load_modules:
            plugins1 = ploader.plugins_all
            self.assertEqual(mock_load_modules.call_count, 1)

            self.assertEqual(len(plugins1.parser), 2)
            self.assertTrue('xml' in plugins1.parser)
            self.assertIsInstance(plugins1.parser.xml, OrderedDict)
            self.assertTrue('json' in plugins1.parser)
            self.assertIsInstance(plugins1.parser.json, OrderedDict)
            self.assertEqual(tuple(plugins1.parser.json.keys()), ('1.0', '2.0'))

            plugins2 = ploader.plugins_all
            self.assertEqual(mock_load_modules.call_count, 1)

            self.assertEqual(plugins1, plugins2)

            # Test for example in docs
            self.assertIs(tuple(plugins1.parser.json.values())[-1], plugins1.parser.json['2.0'])

    def test_blacklist(self):
        """Blacklist prevents listing plugin"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib',
                                      blacklist=[('parser', 'json', '2.0'), ('engine',)])
        plugins = ploader.plugins

        self.assertEqual(len(plugins.engine), 0)
        self.assertEqual(plugins.parser.json.version, '1.0')

    def test_get_plugin(self):
        """Retrieve specific plugin"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        with mock.patch.object(ploader, 'load_modules',
                               wraps=ploader.load_modules) as mock_load_modules:

            jsonplugin = ploader.get_plugin('parser', 'json')
            self.assertEqual(jsonplugin.name, 'json')
            self.assertEqual(jsonplugin.version, '2.0')
            self.assertEqual(mock_load_modules.call_count, 1)

            jsonplugin = ploader.get_plugin('parser', 'json', '1.0')
            self.assertEqual(jsonplugin.name, 'json')
            self.assertEqual(jsonplugin.version, '1.0')
            self.assertEqual(mock_load_modules.call_count, 1)

    def test_get_plugin_missing(self):
        """Attempt to retrieve non-existent plugin, return None"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        self.assertIsNone(ploader.get_plugin('penguin', 'json'))
        self.assertIsNone(ploader.get_plugin('parser', 'penguin'))

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib')
        self.assertIsNone(ploader.get_plugin('parser', 'json', '300.1.1'))

    def test_get_plugin_filtered(self):
        """Retrieve specific plugin if not filtered"""

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib',
                                      type_filter=('engine', 'hook'))

        self.assertIsNone(ploader.get_plugin('parser', 'json'))

        steamplugin = ploader.get_plugin('engine', 'steam')
        self.assertEqual(steamplugin.name, 'steam')

    def test_get_plugin_blacklist(self):
        """Retrieve specific plugin if not blacklisted"""

        blacklist = [('parser', 'json', '2.0'), ('parser', 'xml'), ('engine',)]

        ploader = loader.PluginLoader(group='testdata', library='tests.testdata.lib',
                                      blacklist=blacklist)

        self.assertIsNone(ploader.get_plugin('parser', 'json', '2.0'))
        self.assertIsNone(ploader.get_plugin('parser', 'xml'))
        self.assertIsNone(ploader.get_plugin('engine', 'steam'))

        jsonplugin = ploader.get_plugin('parser', 'json')
        self.assertEqual(jsonplugin.name, 'json')
        self.assertEqual(jsonplugin.version, '1.0')



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Fish hook**
Test data standard library
"""

from __future__ import absolute_import
from tests.testdata.parents import Hook


class Fish(Hook):
    """Dummy fish hook"""

    _alias_ = 'fish'

    def hook(self):
        """Throw hook"""
        return self._alias_



# Copyright 2014 - 2018 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Test module for pluginlib._util**
"""

import sys

import pluginlib._util as util

from tests import TestCase


class TestReraise(TestCase):
    """Tests for raise_with_traceback"""
    def test_raise_with_traceback(self):
        """Caught error should raise as expected"""
        try:
            raise TypeError('Not my type')
        except TypeError as e:
            tback = getattr(e, '__traceback__', sys.exc_info()[2])
            exception = AttributeError('Nah: %s' % str(e))

            with self.assertRaisesRegex(AttributeError, 'Nah: Not my type'):
                util.raise_with_traceback(exception, tback)


class TestClassProperty(TestCase):
    """Tests for ClassProperty decorator class"""

    someproperty = 'Value of a property'

    @util.ClassProperty
    def testproperty(cls):  # noqa: N805 # pylint: disable=no-self-argument
        """Example class property"""
        return cls.someproperty

    def test_class_property(self):
        """Class property should be accessible via standard dot notation"""
        self.assertEqual(self.testproperty, self.someproperty)
        self.assertEqual(self.testproperty, 'Value of a property')


class TestUndefined(TestCase):
    """Tests for Undefined class"""

    def test_str(self):
        """String representation"""
        undef = util.Undefined()
        self.assertEqual(str(undef), 'UNDEF')
        self.assertEqual(repr(undef), 'UNDEF')

    def test_str_nondefault(self):
        """String representation with non-default label"""
        undef = util.Undefined('UNSET')
        self.assertEqual(str(undef), 'UNSET')
        self.assertEqual(repr(undef), 'UNSET')

    def test_bool(self):
        """Should always be False in a Boolean test"""
        undef = util.Undefined()
        self.assertFalse(undef)


class TestAllowBareDecorator(TestCase):
    """Test allow_bare_decorator class decorator decorator"""

    # pylint: disable=no-member
    def test_allow_bare_decorator(self):
        """Decorator should work with or without parentheses"""

        @util.allow_bare_decorator
        class Decorator(object):
            """Sample decorator class"""

            def __init__(self, name=None, group=None):
                self.name = name
                self.group = group

            def __call__(self, cls):
                cls.name = self.name or cls.__name__
                cls.group = self.group or 'default'
                return cls

        @Decorator
        class Class1(object):
            """Bare decorator"""

        self.assertEqual(Class1.__name__, 'Class1')
        self.assertEqual(Class1.name, 'Class1')
        self.assertEqual(Class1.group, 'default')

        @Decorator()
        class Class2(object):
            """Decorator without arguments"""

        self.assertEqual(Class2.__name__, 'Class2')
        self.assertEqual(Class2.name, 'Class2')
        self.assertEqual(Class2.group, 'default')

        @Decorator('spam')
        class Class3(object):
            """Decorator with one argument"""

        self.assertEqual(Class3.__name__, 'Class3')
        self.assertEqual(Class3.name, 'spam')
        self.assertEqual(Class3.group, 'default')

        @Decorator('spam', 'ham')
        class Class4(object):
            """Decorator with two arguments"""

        self.assertEqual(Class4.__name__, 'Class4')
        self.assertEqual(Class4.name, 'spam')
        self.assertEqual(Class4.group, 'ham')


class TestCachingDict(TestCase):
    """Tests for CachingDict dictionary subclass"""
    # pylint: disable=protected-access

    def setUp(self):
        """Stage dictionary and cache"""
        self.cdict = util.CachingDict()
        self.cdict['hello'] = 'world'
        self.cdict._cache['test'] = 'Testing'

    def test_init(self):
        """Standard initialization"""
        self.assertTrue(isinstance(self.cdict._cache, dict))
        self.assertTrue('test' in self.cdict._cache)
        self.cdict._cache.clear()

    def test_setitem(self):
        """Setting a value clears the cache"""
        self.cdict['Hey'] = 'Jude'

    def test_delitem(self):
        """Deleting a value clears the cache"""
        del self.cdict['hello']

    def test_clear(self):
        """Clearing the dictionary also clears the cache"""
        self.cdict.clear()

    def test_setdefault(self):
        """setdefault() only clears the cache when the key doesn't exist"""
        self.assertEqual('world', self.cdict.setdefault('hello', None))
        self.assertTrue('test' in self.cdict._cache)
        self.assertEqual('Jude', self.cdict.setdefault('Hey', 'Jude'))

    def test_pop(self):
        """pop() only clears the cache is the key exists"""
        with self.assertRaises(KeyError):
            self.cdict.pop('Hey')
        self.assertTrue('test' in self.cdict._cache)
        self.assertEqual('world', self.cdict.pop('hello'))

    def test_popitem(self):
        """"popitem() only clears the cache if the dictionary isn't empty"""
        del self.cdict['hello']
        self.cdict._cache['test'] = 'Testing'
        self.assertTrue('test' in self.cdict._cache)

        with self.assertRaises(KeyError):
            self.cdict.popitem()
        self.assertTrue('test' in self.cdict._cache)

        self.setUp()
        self.assertTrue('test' in self.cdict._cache)
        self.assertEqual(('hello', 'world'), self.cdict.popitem())

    def tearDown(self):
        """Ensure cache was cleared"""
        self.assertFalse('test' in self.cdict._cache)
        self.assertFalse(self.cdict._cache)


class TestDictWithDotNotation(TestCase):
    """Tests for DictWithDotNotation dictionary subclass"""

    def test_getter(self):
        """Values can be retrieved by index or dot notation"""
        dotdict = util.DictWithDotNotation({'hello': 'world'})

        self.assertEqual('world', dotdict['hello'])
        self.assertEqual('world', dotdict.hello)

        with self.assertRaises(AttributeError):
            dotdict.notInDict  # pylint: disable=pointless-statement

        with self.assertRaises(KeyError):
            dotdict['notInDict']  # pylint: disable=pointless-statement


class TestClassAbstractStaticMethod(TestCase):
    """Tests for abstractstaticmethod decorator class"""

    def test_abstractstaticmethod(self):
        """Creates a static method marked as an abstract method"""

        def func():
            """Dummy function"""
            return True

        meth = util.abstractstaticmethod(func)
        self.assertIsInstance(meth, staticmethod)
        self.assertTrue(getattr(meth, '__isabstractmethod__', False))
        if sys.version_info[:2] >= (2, 7):
            self.assertTrue(getattr(meth.__func__, '__isabstractmethod__', False))
        else:
            self.assertTrue(getattr(meth.__get__(True), '__isabstractmethod__', False))


class TestClassAbstractClassMethod(TestCase):
    """Tests for abstractclassmethod decorator class"""

    def test_abstractclassmethod(self):
        """Creates a class method marked as an abstract method"""

        def func():
            """Dummy function"""
            return True

        meth = util.abstractclassmethod(func)
        self.assertIsInstance(meth, classmethod)
        self.assertTrue(getattr(meth, '__isabstractmethod__', False))
        if sys.version_info[:2] >= (2, 7):
            self.assertTrue(getattr(meth.__func__, '__isabstractmethod__', False))
        else:
            self.assertTrue(getattr(meth.__get__(True).__func__, '__isabstractmethod__', False))



# Copyright 2017 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
Bad Module that will raise syntax error when imported
"""

# Bad syntax, left off  colon and didn't finish statement
for key in []  # noqa: E999



# Copyright 2018 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib setup file**
"""
import os
import sys

from setuptools import setup, find_packages

from setup_helpers import get_version, readme

INSTALL_REQUIRE = ['setuptools']
TESTS_REQUIRE = []

if sys.version_info[:2] < (3, 3):

    # Include unittest.mock from 3.3+
    TESTS_REQUIRE.append('mock')

if sys.version_info[:2] < (2, 7):

    # Include OrderedDict from 2.7+
    INSTALL_REQUIRE.append('ordereddict')
    # Include unittest features from 2.7+
    TESTS_REQUIRE.append('unittest2')

setup(
    name='pluginlib',
    version=get_version(os.path.join('pluginlib', '__init__.py')),
    description='A framework for creating and importing plugins',
    long_description=readme('README.rst'),
    url='https://github.com/Rockhopper-Technologies/pluginlib',
    license='MPLv2.0',
    author='Avram Lubkin',
    author_email='avylove@rockhopper.net',
    packages=find_packages(exclude=['tests', 'tests.*']),
    install_requires=INSTALL_REQUIRE,
    tests_require=TESTS_REQUIRE,
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='plugin, plugins, pluginlib',
    test_loader="unittest:TestLoader"
)



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**JSON parser**
Test data standard library
"""

from __future__ import absolute_import
from tests.testdata.parents import Parser


class JSON(Parser):
    """Dummy JSON parser"""

    _alias_ = 'json'
    _version_ = '1.0'

    def parse(self):
        return 'json'


class JSON2(Parser):
    """Dummy JSON parser"""

    _alias_ = 'json'
    _version_ = '2.0'

    def parse(self):
        return 'json'



# Copyright 2014 - 2020 Avram Lubkin, All Rights Reserved

# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""
**Pluginlib Object Module**

This module contains pluginlib object classes
"""

from pkg_resources import parse_version

from pluginlib._util import BASESTRING, CachingDict, DictWithDotNotation, OPERATORS, PY26


if PY26:
    from ordereddict import OrderedDict  # pylint: disable=import-error  # pragma: no cover
else:
    from collections import OrderedDict


class BlacklistEntry(object):
    """
    Args:
        plugin_type(str): Parent type
        name(str): Plugin name
        version(str): Plugin version
        operator(str): Comparison operator ('=', '==', '!=', '<', '<=', '>', '>=')

    **Container for blacklist entry**

    If ``operator`` is :py:data:`None` or not specified, it defaults to '=='.

    One of ``plugin_type``, ``name``, or ``version`` must be specified.
    If any are unspecified or :py:data:`None`, they are treated as a wildcard.

    In order to be more compatible with parsed text,
    the order of ``operator`` and ``version`` can be swapped. The following are equivalent:

    .. code-block:: python

        BlacklistEntry('parser', 'json', '1.0', '>=')

    .. code-block:: python

            BlacklistEntry('parser', 'json', '>=', '1.0')

    ``version`` is evaluated using :py:func:`pkg_resources.parse_version`
    and should conform to `PEP 440`_

    .. _PEP 440: https://www.python.org/dev/peps/pep-0440/
    """

    __slots__ = ('type', 'name', 'version', 'operator')

    def __init__(self, plugin_type=None, name=None, version=None, operator=None):

        if plugin_type is name is version is None:
            raise AttributeError('plugin_type, name, or version must be specified')

        self.type = plugin_type
        self.name = name
        if version in OPERATORS:

            self.operator = version
            self.version = operator

            if self.version is None:
                raise AttributeError('version must be specifed when operator is specified')

        else:
            self.version = version
            self.operator = operator

        if self.version is not None and not isinstance(self.version, BASESTRING):
            raise TypeError('version must be a string, recieved %s' % type(self.version).__name__)

        if self.operator is None:
            self.operator = '=='
        elif self.operator not in OPERATORS:
            raise AttributeError("Unsupported operator '%s'" % self.operator)

    def __repr__(self):

        attrs = (self.type, self.name, self.operator, self.version)
        return '%s(%s)' % (self.__class__.__name__, ', '.join([repr(attr) for attr in attrs]))


class GroupDict(DictWithDotNotation):
    """
    Container for a plugin group
    """

    _skip_empty = False
    _key_attr = 'type'
    _bl_skip_attrs = ('name', 'version')
    _bl_empty = DictWithDotNotation

    def _items(self, type_filter=None, name=None):
        """
        Args:
            type_filter(list): Optional iterable of types to return (GroupDict only)
            name(str): Only return key by this name

        Alternative generator for items() method
        """

        if name:
            if type_filter and self._key_attr == 'type':
                if name in type_filter and name in self:
                    yield name, self[name]
            elif name in self:
                yield name, self[name]

        elif type_filter and self._key_attr == 'type':
            for key, val in self.items():
                if key in type_filter:
                    yield key, val
        else:
            for key, val in self.items():
                yield key, val

    def _filter(self, blacklist=None, newest_only=False, type_filter=None, **kwargs):
        """
        Args:
            blacklist(tuple): Iterable of of BlacklistEntry objects
            newest_only(bool): Only the newest version of each plugin is returned
            type(str): Plugin type to retrieve
            name(str): Plugin name to retrieve
            version(str): Plugin version to retrieve

        Returns nested dictionary of plugins

        If a blacklist is supplied, plugins are evaluated against the blacklist entries
        """

        plugins = DictWithDotNotation()
        filtered_name = kwargs.get(self._key_attr, None)

        for key, val in self._items(type_filter, filtered_name):
            plugin_blacklist = None
            skip = False

            if blacklist:

                # Assume blacklist is correct format since it is checked by PluginLoade

                plugin_blacklist = []
                for entry in blacklist:
                    if getattr(entry, self._key_attr) not in (key, None):
                        continue
                    if all(getattr(entry, attr) is None for attr in self._bl_skip_attrs):
                        if not self._skip_empty:
                            plugins[key] = None if filtered_name else self._bl_empty()
                        skip = True
                        break

                    plugin_blacklist.append(entry)

            if not skip:
                # pylint: disable=protected-access
                result = val._filter(plugin_blacklist, newest_only=newest_only, **kwargs)

                if result or not self._skip_empty:
                    plugins[key] = result

        if filtered_name:
            return plugins.get(filtered_name, None)
        return plugins


class TypeDict(GroupDict):
    """
    Container for a plugin type
    """

    _skip_empty = True
    _key_attr = 'name'
    _bl_skip_attrs = ('version',)
    _bl_empty = None  # Not callable, but never called since _skip_empty is True

    def __init__(self, parent, *args, **kwargs):
        super(TypeDict, self).__init__(*args, **kwargs)
        self._parent = parent


class PluginDict(CachingDict):
    """
    Dictionary with properties for retrieving plugins
    """

    def _sorted_keys(self):
        """
        Return list of keys sorted by version

        Sorting is done based on :py:func:`pkg_resources.parse_version`
        """

        try:
            keys = self._cache['sorted_keys']
        except KeyError:
            keys = self._cache['sorted_keys'] = sorted(self.keys(), key=parse_version)

        return keys

    def _process_blacklist(self, blacklist):
        """
        Process blacklist into set of excluded versions
        """

        # Assume blacklist is correct format since it is checked by PluginLoader

        blacklist_cache = {}
        blacklist_cache_old = self._cache.get('blacklist', {})

        for entry in blacklist:

            blackkey = (entry.version, entry.operator)

            if blackkey in blacklist_cache:
                continue

            if blackkey in blacklist_cache_old:
                blacklist_cache[blackkey] = blacklist_cache_old[blackkey]
            else:
                entry_cache = blacklist_cache[blackkey] = set()
                blackversion = parse_version(entry.version or '0')
                blackop = OPERATORS[entry.operator]

                for key in self:
                    if blackop(parse_version(key), blackversion):
                        entry_cache.add(key)

        self._cache['blacklist'] = blacklist_cache
        return set().union(*blacklist_cache.values())

    def _filter(self, blacklist=None, newest_only=False, **kwargs):
        """
        Args:
            blacklist(tuple): Iterable of of BlacklistEntry objects
            newest_only(bool): Only the newest version of each plugin is returned
            version(str): Specific version to retrieve

        Returns dictionary of plugins

        If a blacklist is supplied, plugins are evaluated against the blacklist entries
        """

        version = kwargs.get('version', None)
        rtn = None

        if self:  # Dict is not empty

            if blacklist:

                blacklist = self._process_blacklist(blacklist)

                if version:
                    if version not in blacklist:
                        rtn = self.get(version, None)

                elif newest_only:
                    for key in reversed(self._sorted_keys()):
                        if key not in blacklist:
                            rtn = self[key]
                            break
                    # If no keys are left, None will be returned
                else:
                    rtn = OrderedDict()
                    for key in self._sorted_keys():
                        if key not in blacklist:
                            rtn[key] = self[key]

            elif version:
                rtn = self.get(version, None)

            elif newest_only:
                rtn = self[self._sorted_keys()[-1]]

            else:
                rtn = OrderedDict((key, self[key]) for key in self._sorted_keys())

        return rtn
